import numpy as np
from scipy import stats

from pimp.epm.base_epm import RandomForestWithInstances as rfi

__author__ = "Andre Biedenkapp"
__copyright__ = "Copyright 2016, ML4AAD"
__license__ = "3-clause BSD"
__maintainer__ = "Andre Biedenkapp"
__email__ = "biedenka@cs.uni-freiburg.de"


class Unloggedrfwi(rfi):

    def __init__(self, configspace, types, bounds, seed, **kwargs):
        """
        Interface to the random forest that takes instance features
        into account.

        Parameters
        ----------
        configspace: ConfigurationSpace
            configspace to be passed to random forest (used to impute inactive parameter-values)
        types: np.ndarray (D)
            Specifies the number of categorical values of an input dimension. Where
            the i-th entry corresponds to the i-th input dimension. Let say we have
            2 dimension where the first dimension consists of 3 different
            categorical choices and the second dimension is continuous than we
            have to pass np.array([2, 0]). Note that we count starting from 0.
        bounds: np.ndarray (D)
            Specifies the bounds
        seed: int
            The seed that is passed to the random_forest_run library.

        instance_features: np.ndarray (I, K)
            Contains the K dimensional instance features
            of the I different instances
        num_trees: int
            The number of trees in the random forest.
        do_bootstrapping: bool
            Turns on / off bootstrapping in the random forest.
        ratio_features: float
            The ratio of features that are considered for splitting.
        min_samples_split: int
            The minimum number of data points to perform a split.
        min_samples_leaf: int
            The minimum number of data points in a leaf.
        max_depth: int

        eps_purity: float

        max_num_nodes: int

        cutoff: int
            The cutoff used in the specified scenario

        threshold:
            Maximal possible value
        """
        super().__init__(configspace=configspace, types=types, bounds=bounds, seed=seed, **kwargs)

    # With the usage of pyrfr 0.8.0 this method is obsolete.
    # def _predict(self, X):
    #     """Predict means and variances for given X by first unlogging the leaf-values and then computing the mean for
    #     the trees NOT the training batch afterwards. The mean for the whole batch is handled by the parent class!
    #
    #     Parameters
    #     ----------
    #     X : np.ndarray of shape = [n_samples, n_features (config + instance
    #     features)]
    #
    #     Returns
    #     -------
    #     means : np.ndarray of shape = [n_samples, 1]
    #         Predictive mean
    #     vars : np.ndarray  of shape = [n_samples, 1]
    #         Predictive variance
    #     """
    #     if len(X.shape) != 2:
    #         raise ValueError(
    #             'Expected 2d array, got %dd array!' % len(X.shape))
    #     if X.shape[1] != self.types.shape[0]:
    #         raise ValueError('Rows in X should have %d entries but have %d!' %
    #                          (self.types.shape[0], X.shape[1]))
    #
    #     tree_mean_predictions = []
    #     tree_mean_variances = []
    #     for x in X:
    #         tmpx = np.array(list(map(lambda x_: np.power(10, x_), self.rf.all_leaf_values(x))))  # unlog values
    #         tree_mean_predictions.append(list(map(lambda x_: np.mean(x_), tmpx)))  # calculate mean and var
    #         tree_mean_variances.append(list(map(lambda x_: np.var(x_), tmpx)))  # over individual trees
    #
    #     mean = np.mean(tree_mean_predictions, axis=1)
    #     var = np.mean(tree_mean_variances, axis=1)
    #
    #     return mean.reshape((-1, 1)), var.reshape((-1, 1))

    def predict(self, X):
        """
        Method to override the predict method of RandomForestWithInstances.
        Thus it can be used in the marginalized over instances method of the RFWI class
        """
        return self._predict(X)



import os
from collections import OrderedDict
from copy import deepcopy

import numpy as np
from bokeh.models import Row, Panel, Tabs
from tqdm import tqdm

from pimp.utils.bokeh_helpers import bokeh_boxplot, bokeh_line_uncertainty, save_and_show

tqdm.monitor_interval = 0
import matplotlib as mpl
mpl.use('Agg')
from matplotlib import pyplot as plt

from pimp.configspace import change_hp_value, Configuration, ForbiddenValueError,\
    impute_inactive_values, CategoricalHyperparameter, NumericalHyperparameter, check_forbidden
from pimp.evaluator.base_evaluator import AbstractEvaluator

__author__ = "Andre Biedenkapp"
__copyright__ = "Copyright 2017, ML4AAD"
__license__ = "3-clause BSD"
__maintainer__ = "Andre Biedenkapp"
__email__ = "biedenka@cs.uni-freiburg.de"


class LPI(AbstractEvaluator):

    """
    Implementation of Ablation via surrogates
    """

    def __init__(self, scenario, cs, model, to_evaluate: int, incumbent=None, continous_neighbors=500,
                 old_sampling=False, show_query_points=False, quant_var=True, **kwargs):
        super().__init__(scenario, cs, model, to_evaluate, **kwargs)
        self.name = 'LPI'
        self.logger = 'pimp.' + self.name
        self.incumbent = incumbent
        self.incumbent_dict = self.incumbent.get_dictionary()
        self._continous_param_neighbor_samples = continous_neighbors
        self.show_query_points = show_query_points
        self.old_sampling = old_sampling
        self.neighborhood_dict = None
        self.performance_dict = {}
        self._sampled_neighbors = 0
        self.variance_dict = {}
        self.quantify_importance_via_variance = quant_var
        self.evaluated_parameter_importance_uncertainty = OrderedDict()

    def _old_sampling_of_one_exchange_neighborhood(self, param, array, index):
        neighbourhood = []
        number_of_sampled_neighbors = 0
        iteration = 0
        checked_neighbors = []
        checked_neighbors_non_unit_cube = []
        while True:
            hp = self.incumbent.configuration_space.get_hyperparameter(param)
            num_neighbors = hp.get_num_neighbors(self.incumbent.get(param))
            self.logger.debug('\t' + str(num_neighbors))

            # Obtain neigbors differently for different possible numbers of
            # neighbors
            if num_neighbors == 0:
                self.logger.debug('\tNo neighbors!')
                break
            # No infinite loops
            elif iteration > 500:
                self.logger.debug('\tMax iter')
                break
            elif np.isinf(num_neighbors):
                num_samples_to_go = min(10, self._continous_param_neighbor_samples - number_of_sampled_neighbors)
                if number_of_sampled_neighbors >= self._continous_param_neighbor_samples or num_samples_to_go <= 0:
                    break
                neighbors = hp.get_neighbors(array[index], self.rng,
                                             number=num_samples_to_go)
            else:
                if iteration > 0:
                    break
                neighbors = hp.get_neighbors(array[index], self.rng)
            # self.logger.debug('\t\t' + str(neighbors))
            # Check all newly obtained neighbors
            for neighbor in neighbors:
                if neighbor in checked_neighbors:
                    iteration += 1
                    continue
                new_array = array.copy()
                new_array = change_hp_value(self.incumbent.configuration_space, new_array, param, neighbor,
                                            index)
                try:
                    new_configuration = Configuration(self.incumbent.configuration_space, vector=new_array)
                    neighbourhood.append(new_configuration)
                    new_configuration.is_valid_configuration()
                    check_forbidden(self.cs.forbidden_clauses, new_array)
                    number_of_sampled_neighbors += 1
                    checked_neighbors.append(neighbor)
                    checked_neighbors_non_unit_cube.append(new_configuration[param])
                except (ForbiddenValueError, ValueError) as e:
                    pass
                iteration += 1
        return checked_neighbors, checked_neighbors_non_unit_cube

    def _get_one_exchange_neighborhood_by_parameter(self):
        """
        Slight modification of ConfigSpace's get_one_exchange neighborhood. This orders the parameter values and samples
        more neighbors in one go. Further we need to rigorously check each and every neighbor if it is forbidden or not.
        """
        neighborhood_dict = {}
        params = list(self.incumbent.keys())
        self.logger.debug('params: ' + str(params))
        for index, param in enumerate(params):
            self.logger.info('Sampling neighborhood of %s' % param)
            array = self.incumbent.get_array()

            if not np.isfinite(array[index]):
                self.logger.info('>'.join(['-'*50, ' Not active!']))
                continue
            if self.old_sampling:
                checked_neighbors, checked_neighbors_non_unit_cube = self._old_sampling_of_one_exchange_neighborhood(
                    param, array, index
                )
            else:
                neighbourhood = []
                checked_neighbors = []
                checked_neighbors_non_unit_cube = []
                hp = self.incumbent.configuration_space.get_hyperparameter(param)
                num_neighbors = hp.get_num_neighbors(self.incumbent.get(param))
                self.logger.debug('\t' + str(num_neighbors))
                if num_neighbors == 0:
                    self.logger.debug('\tNo neighbors!')
                    continue
                elif np.isinf(num_neighbors):  # Continous Parameters
                    if hp.log:
                        base = np.e
                        log_lower = np.log(hp.lower) / np.log(base)
                        log_upper = np.log(hp.upper) / np.log(base)
                        neighbors = np.logspace(log_lower, log_upper, self._continous_param_neighbor_samples,
                                                endpoint=True, base=base)
                    else:
                        neighbors = np.linspace(hp.lower, hp.upper, self._continous_param_neighbor_samples)
                    neighbors = list(map(lambda x: hp._inverse_transform(x), neighbors))
                else:
                    neighbors = hp.get_neighbors(array[index], self.rng)
                for neighbor in neighbors:
                    if neighbor in checked_neighbors:
                        continue
                    new_array = array.copy()
                    new_array = change_hp_value(self.incumbent.configuration_space, new_array, param, neighbor,
                                                index)
                    try:
                        new_configuration = Configuration(self.incumbent.configuration_space, vector=new_array)
                        neighbourhood.append(new_configuration)
                        new_configuration.is_valid_configuration()
                        check_forbidden(self.cs.forbidden_clauses, new_array)
                        checked_neighbors.append(neighbor)
                        checked_neighbors_non_unit_cube.append(new_configuration[param])
                    except (ForbiddenValueError, ValueError) as e:
                        pass
            self.logger.info('>'.join(['-'*50, ' Found {:>3d} valid neighbors'.format(len(checked_neighbors))]))
            self._sampled_neighbors += len(checked_neighbors) + 1
            sort_idx = list(map(lambda x: x[0], sorted(enumerate(checked_neighbors), key=lambda y: y[1])))
            if isinstance(self.cs.get_hyperparameter(param), CategoricalHyperparameter):
                checked_neighbors_non_unit_cube = list(np.array(checked_neighbors_non_unit_cube)[sort_idx])
            else:
                checked_neighbors_non_unit_cube = np.array(checked_neighbors_non_unit_cube)[sort_idx]
            neighborhood_dict[param] = [np.array(checked_neighbors)[sort_idx], checked_neighbors_non_unit_cube]
        return neighborhood_dict

    def run(self) -> OrderedDict:
        """
        Main function.

        Returns
        -------
        evaluated_parameter_importance:OrderedDict
            Parameter -> importance. The order is important as smaller indices indicate higher importance
        """
        neighborhood_dict = self._get_one_exchange_neighborhood_by_parameter()  # sampled on a unit-hypercube!
        self.neighborhood_dict = neighborhood_dict
        incumbent_array = self.incumbent.get_array()
        def_perf, def_var = self._predict_over_instance_set(impute_inactive_values(self.cs.get_default_configuration()))
        inc_perf, inc_var = self._predict_over_instance_set(impute_inactive_values(self.incumbent))
        delta = def_perf - inc_perf
        evaluated_parameter_importance = {}

        # These are used for plotting and hold the predictions for each neighbor of each parameter
        # That means performance_dict holds the mean, variance_dict the variance of the forest
        performance_dict = {}
        variance_dict = {}
        # This are used for importance and hold the corresponding importance/variance over neighbors
        # Only import if NOT quantifying importance via performance-variance across neighbours
        overall_imp = {}
        # Nested list of values per tree in random forest
        pred_per_tree = {}

        # Iterate over parameters
        pbar = tqdm(range(self._sampled_neighbors), ascii=True, disable=not self.verbose)
        for index, param in enumerate(self.incumbent.keys()):
            if param not in neighborhood_dict:
                pbar.set_description('{: >.70s}'.format('Parameter %s is inactive' % param))
                continue

            pbar.set_description('Predicting performances for neighbors of {: >.30s}'.format(param))
            performance_dict[param] = []
            variance_dict[param] = []
            pred_per_tree[param] = []
            added_inc = False
            inc_at = 0
            # Iterate over neighbors
            for unit_neighbor, neighbor in zip(neighborhood_dict[param][0], neighborhood_dict[param][1]):
                if not added_inc:
                    # Detect incumbent
                    if unit_neighbor > incumbent_array[index]:
                        performance_dict[param].append(inc_perf)
                        variance_dict[param].append(inc_var)
                        pbar.update(1)
                        added_inc = True
                    else:
                        inc_at += 1
                # self.logger.debug('%s -> %s' % (self.incumbent[param], neighbor))
                # Create the neighbor-Configuration object
                new_array = incumbent_array.copy()
                new_array = change_hp_value(self.incumbent.configuration_space, new_array,
                                            param, unit_neighbor, index)
                new_configuration = impute_inactive_values(Configuration(self.incumbent.configuration_space,
                                                                         vector=new_array))
                # Predict performance
                x = np.array(new_configuration.get_array())
                pred_per_tree[param].append([np.mean(tree_pred) for tree_pred in self.model.rf.all_leaf_values(x)])
                # self.logger.debug("Pred per tree: %s", str(pred_per_tree[param][-1]))
                performance_dict[param].append(np.mean(pred_per_tree[param][-1]))
                variance_dict[param].append(np.var(pred_per_tree[param][-1]))

                pbar.update(1)
            if len(neighborhood_dict[param][0]) > 0:
                neighborhood_dict[param][0] = np.insert(neighborhood_dict[param][0], inc_at, incumbent_array[index])
                neighborhood_dict[param][1] = np.insert(neighborhood_dict[param][1], inc_at, self.incumbent[param])
            else:
                neighborhood_dict[param][0] = np.array(incumbent_array[index])
                neighborhood_dict[param][1] = [self.incumbent[param]]
            if not added_inc:
                mean, var = self._predict_over_instance_set(impute_inactive_values(self.incumbent))
                performance_dict[param].append(mean)
                variance_dict[param].append(var)
                pbar.update(1)
            # After all neighbors are estimated, look at all performances except the incumbent
            tmp_perf = performance_dict[param][:inc_at] + performance_dict[param][inc_at + 1:]
            if delta == 0:
                delta = 1  # To avoid division by zero
            imp_over_mea = (np.mean(tmp_perf) - performance_dict[param][inc_at]) / delta
            imp_over_med = (np.median(tmp_perf) - performance_dict[param][inc_at]) / delta
            try:
                imp_over_max = (np.max(tmp_perf) - performance_dict[param][inc_at]) / delta
            except ValueError:
                imp_over_max = np.nan  # Hacky fix as this is never used anyway
            overall_imp[param] = np.array([imp_over_mea, imp_over_med, imp_over_max])

        # Creating actual importance value (by normalizing over sum of vars)
        num_trees = len(list(pred_per_tree.values())[0][0])
        params = list(performance_dict.keys())
        overall_var_per_tree = {
                param : [np.var([neighbor[tree_idx] for neighbor in pred_per_tree[param]])
                            for tree_idx in range(num_trees)]
                                for param in params}
        # Sum up variances per tree across parameters
        sum_var_per_tree = [sum([overall_var_per_tree[param][tree_idx] for param in params])
                                     for tree_idx in range(num_trees)]
        # Normalize
        overall_var_per_tree = {p : [t / sum_var_per_tree[idx] for idx, t in enumerate(trees)] for p, trees in
                                                    overall_var_per_tree.items()}
        self.logger.debug("overall_var_per_tree %s (%d trees)",
                str(overall_var_per_tree), len(list(pred_per_tree.values())[0][0]))
        self.logger.debug("sum_var_per_tree %s (%d trees)",
                str(sum_var_per_tree), len(list(pred_per_tree.values())[0][0]))
        for param in performance_dict.keys():
            if self.quantify_importance_via_variance:
                evaluated_parameter_importance[param] = np.mean(overall_var_per_tree[param])
            else:
                evaluated_parameter_importance[param] = overall_imp[param][0]

        only_show = sorted(list(evaluated_parameter_importance.keys()),
                           key=lambda p: evaluated_parameter_importance[p])[:min(self.to_evaluate,
                                                                                 len(evaluated_parameter_importance.keys()))]

        self.neighborhood_dict = neighborhood_dict
        self.performance_dict = performance_dict
        self.variance_dict = variance_dict
        self.evaluated_parameter_importance = OrderedDict([(p, evaluated_parameter_importance[p]) for p in only_show])
        if self.quantify_importance_via_variance:
            self.evaluated_parameter_importance_uncertainty = OrderedDict([(p, np.std(overall_var_per_tree[p])) for p in only_show])
        all_res = {'imp': self.evaluated_parameter_importance,
                   'order': list(self.evaluated_parameter_importance.keys())}
        return all_res

    def _predict_over_instance_set(self, config):
        """
        Small wrapper to predict marginalized over instances
        Parameter
        ---------
        config:Configuration
            The self.incumbent of wich the performance across the whole instance set is to be estimated
        Returns
        -------
        mean
            the mean performance over the instance set
        var
            the variance over the instance set. If logged values are used, the variance might not be able to be used
        """
        mean, var = self.model.predict_marginalized_over_instances(np.array([config.get_array()]))
        return mean.squeeze(), var.squeeze()

    def plot_result(self, name='incneighbor', show=True):
        if not os.path.exists(name):
            os.mkdir(name)
        keys = deepcopy(list(self.incumbent.keys()))
        pbar = tqdm(list(keys), ascii=True, disable=not self.verbose)
        y_label = self.scenario.run_obj if self.scenario.run_obj != 'quality' else 'cost'
        for param in pbar:
            pbar.set_description('Plotting results for %s' % param)
            if param in self.performance_dict:
                fig = plt.figure()
                ax1 = fig.add_subplot(111)
                plt.subplots_adjust(bottom=0.25, top=0.9, left=0.05, right=.95)
                p, v = self.performance_dict[param], self.variance_dict[param]

                std = np.array(list(map(lambda s: np.sqrt(s), v))).flatten()
                upper = np.array(list(map(lambda x, y: x + y, p, std))).flatten()
                if self.scenario.run_obj == "runtime":
                    lower = np.array(list(map(lambda x, y: max(x - y, 0), p, std))).squeeze()
                else:
                    lower = np.array(list(map(lambda x, y: x - y, p, std))).squeeze()
                min_y = min(lower)
                max_y = max(upper)
                if not isinstance(self.incumbent.configuration_space.get_hyperparameter(param),
                                  CategoricalHyperparameter):
                    ax1.fill_between(self.neighborhood_dict[param][1], lower, upper, label='std', color=self.area_color)
                    ax1.plot(self.neighborhood_dict[param][1], p, label='predicted %s' % y_label,
                             ls='-', zorder=80,
                             **self.LINE_FONT)
                    label = True
                    c_inc = True
                    for n_idx, neighbor in enumerate(self.neighborhood_dict[param][1]):
                        if neighbor == self.incumbent[param] and c_inc:
                            ax1.scatter(neighbor, p[n_idx], label='incumbent', c='r', marker='.', zorder=999)
                            c_inc = False
                        elif self.show_query_points:
                            if label:
                                ax1.scatter(neighbor, p[n_idx], label='query points', c='w', marker='.',
                                            zorder=100, edgecolors='k')
                                label = False
                            else:
                                ax1.scatter(neighbor, p[n_idx], c='w', marker='.', zorder=100, edgecolors='k')
                    ax1.xaxis.grid(True)
                    ax1.yaxis.grid(True)
                    ax1.legend()
                else:
                    b = ax1.boxplot([[x] for x in p], showfliers=False)
                    plt.xticks(np.arange(1, len(self.neighborhood_dict[param][1])+1, 1),
                               self.neighborhood_dict[param][1])
                    # blow up boxes
                    for box, std_ in zip(b["boxes"], std):
                        y = box.get_ydata()
                        y[2:4] = y[2:4] + std_
                        if self.scenario.run_obj == "runtime":
                            y[0] = max(y[0] - std_, 0)
                            y[1] = max(y[1] - std_, 0)
                            y[4] = max(y[4] - std_, 0)
                        else:
                            y[0:2] = y[0:2] - std_
                            y[4] = y[4] - std_
                        box.set_ydata(y)
                        if self.scenario.run_obj == "runtime":
                            min_y = min(min_y, max(y[0] - std_, -0.01))
                        else:
                            min_y = min(min_y, y[0] - std_)
                        max_y = max(max_y, y[2] + std_)
                    for idx, t in enumerate(ax1.xaxis.get_ticklabels()):
                        color_ = (0., 0., 0.)
                        if self.neighborhood_dict[param][1][idx] == self.incumbent[param]:
                            color_ = (1, 0, 0)
                        t.set_color(color_)

                if isinstance(self.cs.get_hyperparameter(param), NumericalHyperparameter):
                    ax1.set_xscale('log' if self.cs.get_hyperparameter(param).log else 'linear')
                plt.xlabel(param)
                if self.scenario.run_obj == 'runtime':
                    ax1.set_ylabel('runtime [sec]', zorder=81, **self.LABEL_FONT)
                else:
                    ax1.set_ylabel('%s' % y_label, zorder=81, **self.LABEL_FONT)
                try:
                    plt.tight_layout()
                except ValueError:
                    pass
                ax1.set_ylim([min_y * 0.95, max_y])
                plt.savefig(os.path.join(name, param + '.png'))
                ax1.set_yscale('log')
                if min_y <= 0:
                    min_y = list(filter(lambda x: x > 0, lower))
                    if min_y:
                        min_y = min(min_y)
                    else:
                        min_y = 10**-3
                ax1.set_ylim([min_y * 0.95, max_y])
                plt.savefig(os.path.join(name, param + '_log.png'))
                plt.close('all')

    def plot_bokeh(self, plot_name=None, show_plot=False):
        """
        Plot local parameter importance in an interactive bokeh-plot.

        Parameters
        ----------
        plot_name: str
            path where to store the plot, None to not save it
        show_plot: bool
            whether or not to open plot in standard browser

        Returns
        -------
        layout: bokeh.models.Row
            bokeh plot (can be used in notebook or comparted with components)
        """
        plots = []
        pbar = tqdm(deepcopy(list(self.incumbent.keys())), ascii=True, disable=not self.verbose)
        for param in pbar:
            pbar.set_description('Plotting results (in bokeh) for %s' % param)
            if param in self.performance_dict:
                # Data to be plotted:
                p, v = self.performance_dict[param], self.variance_dict[param]
                std = np.array(list(map(lambda s: np.sqrt(s), v))).flatten()

                # Find incumbent indices
                inc_indices = [n_idx for n_idx, neighbor in enumerate(self.neighborhood_dict[param][1])
                               if neighbor == self.incumbent[param]]

                # Boxplot if categorical, else line
                if isinstance(self.incumbent.configuration_space.get_hyperparameter(param), CategoricalHyperparameter):
                    p = bokeh_boxplot(self.neighborhood_dict[param][1], p, std,
                                      y_label="runtime [sec]" if self.scenario.run_obj == "runtime" else "cost",
                                      x_label=param,
                                      runtime=self.scenario.run_obj=="runtime",
                                      inc_indices=inc_indices)
                else:
                    p = bokeh_line_uncertainty(self.neighborhood_dict[param][1], p, std,
                                               self.cs.get_hyperparameter(param).log,
                                               y_label="runtime [sec]" if self.scenario.run_obj == "runtime" else "cost",
                                               x_label=param,
                                               inc_indices=inc_indices)

            # Add as Panel so it can be easily put into Tabs
            plots.append(Panel(child=Row(p), title=param))

        # Putting the Tabs together
        layout = Tabs(tabs=[*plots])

        save_and_show(plot_name, show_plot, layout)

        return layout



import logging
import argcomplete
import re as _re
import os
from argparse import SUPPRESS, ArgumentDefaultsHelpFormatter, ArgumentParser, HelpFormatter
from argparse import OPTIONAL, ZERO_OR_MORE, ONE_OR_MORE, REMAINDER, PARSER


from pimp.__version__ import __version__ as v

__author__ = "Andre Biedenkapp"
__copyright__ = "Copyright 2016, ML4AAD"
__license__ = "3-clause BSD"
__maintainer__ = "Andre Biedenkapp"
__email__ = "biedenka@cs.uni-freiburg.de"


############################ CUSTUMIZATION OF ARGPARSE OUTPUT #######################################
try:
    from gettext import gettext as _, ngettext
except ImportError:
    def _(message):
        return message

    def ngettext(singular,plural,n):
        if n == 1:
            return singular
        else:
            return plural


class SmartArgsDefHelpFormatter(ArgumentDefaultsHelpFormatter):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._max_help_position = 40
        self._width = 120

    def _get_help_string(self, action):
        help = action.help
        if '%(default)' not in action.help:
            if action.default is not SUPPRESS and action.nargs is None:
                help += '(default: %(default)s)'
        return help

    def _split_lines(self, text, width):
        if text.startswith('raw|'):
            return text[4:].splitlines()
        return HelpFormatter._split_lines(self, text, width)

    def _format_args(self, action, default_metavar):
        """
        Adaptation of argparse _format_args to ignore special output for variable nargs
        """
        get_metavar = self._metavar_formatter(action, default_metavar)
        if action.nargs is None:
            result = '%s' % get_metavar(1)
        elif action.nargs == OPTIONAL:
            result = '[%s]' % get_metavar(1)
        elif action.nargs == ZERO_OR_MORE:
            result = '%s' % get_metavar(1)  # '[%s [%s ...]]' % get_metavar(2)
        elif action.nargs == ONE_OR_MORE:
            result = '%s' % get_metavar(1)  # '%s [%s ...]' % get_metavar(2)
        elif action.nargs == REMAINDER:
            result = '...'
        elif action.nargs == PARSER:
            result = '%s ...' % get_metavar(1)
        else:
            formats = ['%s' for _ in range(action.nargs)]
            result = ' '.join(formats) % get_metavar(action.nargs)
        return result

    def _metavar_formatter(self, action, default_metavar):
        if action.metavar is not None:
            result = action.metavar
        elif action.choices is not None:
            choices = [action.choices[0], action.choices[-1]]
            choice_strs = [str(choice) for choice in choices]
            if len(action.choices) <= 2:
                result = '{%s}' % ', '.join(choice_strs)
            else:
                result = '{%s}' % '{}, ..., {}'.format(*choice_strs)
        else:
            result = default_metavar

        def format(tuple_size):
            if isinstance(result, tuple):
                return result
            else:
                return (result, ) * tuple_size
        return format

    def _format_usage(self, usage, actions, groups, prefix):
        """
        Nearly identical to argparses _format usage except suppressing an assert to allow for custom choice behaviour
        """
        if prefix is None:
            prefix = _('usage: ')

        # if usage is specified, use that
        if usage is not None:
            usage = usage % dict(prog=self._prog)

        # if no optionals or positionals are available, usage is just prog
        elif usage is None and not actions:
            usage = '%(prog)s' % dict(prog=self._prog)

        # if optionals and positionals are available, calculate usage
        elif usage is None:
            prog = '%(prog)s' % dict(prog=self._prog)

            # split optionals from positionals
            optionals = []
            positionals = []
            for action in actions:
                if action.option_strings:
                    optionals.append(action)
                else:
                    positionals.append(action)

            # build full usage string
            format = self._format_actions_usage
            action_usage = format(optionals + positionals, groups)
            usage = ' '.join([s for s in [prog, action_usage] if s])

            # wrap the usage parts if it's too long
            text_width = self._width - self._current_indent
            if len(prefix) + len(usage) > text_width:

                # break usage into wrappable parts
                part_regexp = r'\(.*?\)+|\[.*?\]+|\S+'
                opt_usage = format(optionals, groups)
                pos_usage = format(positionals, groups)
                opt_parts = _re.findall(part_regexp, opt_usage)
                pos_parts = _re.findall(part_regexp, pos_usage)
                # assert ' '.join(opt_parts) == opt_usage
                assert ' '.join(pos_parts) == pos_usage

                # helper for wrapping lines
                def get_lines(parts, indent, prefix=None):
                    lines = []
                    line = []
                    if prefix is not None:
                        line_len = len(prefix) - 1
                    else:
                        line_len = len(indent) - 1
                    for part in parts:
                        if line_len + 1 + len(part) > text_width and line:
                            lines.append(indent + ' '.join(line))
                            line = []
                            line_len = len(indent) - 1
                        line.append(part)
                        line_len += len(part) + 1
                    if line:
                        lines.append(indent + ' '.join(line))
                    if prefix is not None:
                        lines[0] = lines[0][len(indent):]
                    return lines

                # if prog is short, follow it with optionals or positionals
                if len(prefix) + len(prog) <= 0.75 * text_width:
                    indent = ' ' * (len(prefix) + len(prog) + 1)
                    if opt_parts:
                        lines = get_lines([prog] + opt_parts, indent, prefix)
                        lines.extend(get_lines(pos_parts, indent))
                    elif pos_parts:
                        lines = get_lines([prog] + pos_parts, indent, prefix)
                    else:
                        lines = [prog]

                # if prog is long, put it on its own line
                else:
                    indent = ' ' * len(prefix)
                    parts = opt_parts + pos_parts
                    lines = get_lines(parts, indent)
                    if len(lines) > 1:
                        lines = []
                        lines.extend(get_lines(opt_parts, indent))
                        lines.extend(get_lines(pos_parts, indent))
                    lines = [prog] + lines

                # join lines into usage
                usage = '\n'.join(lines)

        # prefix with 'usage:'
        return '%s%s\n\n' % (prefix, usage)


##################################### ACTUAL PARSER SETUP ##############################
class CMDs:
    """
        use argparse to parse command line options

        Attributes
        ----------
        logger : Logger oject
    """

    def read_cmd(self):
        """
            reads command line options

            Returns
            -------
                args_: parsed arguments; return of parse_args of ArgumentParser
        """
        global v
        m_choices = ['ablation',
                     'all',
                     'fanova',
                     'forward-selection',
                     'incneighbor',
                     'influence-model',
                     'lpi']
        parser = ArgumentParser(formatter_class=SmartArgsDefHelpFormatter,
                                description='%s' % (
                                    "PIMP implements a multitude of "
                                    "parameter importance methods to determine which parameters have the most "
                                    "influence over the algorithms behaviour, given data from an algorithm configurator"
                                    "such as SMAC3. Version " + str(v)),
                                add_help=False)
        req_opts = parser.add_argument_group("required arguments:" + '~'*100)
        req_opts.add_argument("-S", "--scenario_file",
                              required=True,
                              help="(path to) scenario file in AClib format.",
                              default=SUPPRESS)
        req_opts.add_argument("-M", "--modus",
                              required=True,
                              nargs='+',
                              choices=m_choices,
                              help='analysis method(s) to use. Choose any combination from \n{' +
                                   ', '.join(m_choices) + '}',
                              type=str.lower,
                              default=SUPPRESS)
        req_opts.add_argument("-H", "--history",
                              required=True,
                              help="(path to) runhistory file that contains all "
                                   "collected data from an optimization run",
                              default=SUPPRESS)

        opt_opts = parser.add_argument_group("optional arguments:" + '~'*100)
        opt_opts.add_argument("--seed",
                              default=12345,
                              type=int,
                              help="raw|random seed. Used internally when fitting the random forest.\n")
        opt_opts.add_argument("-V", "--verbose_level",
                              default=logging.INFO,
                              choices=["INFO", "DEBUG"],
                              help="raw|verbosity\n")
        opt_opts.add_argument("-T", "--trajectory",
                              default=None,
                              help="raw|(path to) trajectory file. Needs to contain at least one line detailing the\n"
                                   "best seen configuration and it's cost.\n")
        opt_opts.add_argument("-N", "--num_params",
                              default=-1,
                              type=int,
                              help="raw|number of parameters to evaluate. -1 -> use all\n")
        opt_opts.add_argument("-P", "--max_sample_size",
                              default=-1,
                              type=int,
                              help="raw|number of samples from runhistorie(s) used. -1 -> use all\n")
        opt_opts.add_argument('-F', '--out-folder',
                              default=None,
                              help='raw|folder to store all results in\n',
                              dest='out_folder')
        opt_opts.add_argument('-D', '--working_dir',
                              default='.',
                              help='raw|working directory. Contains all necessary files such as scenario, (path to)\n'
                                   'runhistory, [(path to) trajectory].\n',
                              dest='wdir')
        opt_opts.add_argument("-I", "--impute",
                              action='store_true',
                              help="If set, censored data will be imputed.")
        opt_opts.add_argument("-C", "--table",
                              action='store_true',
                              help="=> a .tex file will be created containing all results in an easy to compare"
                                   "fashion. Parameters will be sorted by importance according to the first method"
                                   "in the table.")
        opt_opts.add_argument('--fanova_cut_at_default',
                              action='store_true',
                              help='cut fANOVA results at the default. This quantifies importance only in'
                                   ' terms of improvement over the'
                                   ' default.')
        opt_opts.add_argument('--fanova_no_pairs',
                              action='store_false',
                              help="=> fANOVA won't compute pairwise marginals",
                              dest='fanova_pairwise')
        opt_opts.add_argument('--lpi_quantify_perf_improvement',
                              action='store_false',
                              help="=> LPI computes importance as performance improvement",
                              dest='incn_quant_var')
        opt_opts.add_argument('--forward_sel_feat_imp',
                              action='store_true',
                              help="=> forward selection computes feature importance, not parameter importance",
                              dest='forwardsel_feat_imp')
        opt_opts.add_argument('--forwardsel_cv',
                              action='store_true',
                              help='=> forward selection errors computed via cross-validation instead of OOBs',
                              dest='forwardsel_cv')
        opt_opts.add_argument('--marginalize_over_instances',
                              action='store_true',
                              help='=> deactivate preprocessing step in which instances are marginalized '
                                   'away to speedup'
                                   ' ablation, forward-selection and LPI predictions',
                              dest='marg_inst')
        spe_opts = parser.add_argument_group("special arguments:" + '~'*100)
        spe_opts.add_argument('-v', '--version', action='version',
                              version='%(prog)s ' + str(v), help="show program's version number and exit.")
        spe_opts.add_argument("-h", "--help", action="help", help="show this help message and exit")
        argcomplete.autocomplete(parser)
        args_, misc = parser.parse_known_args()
        self._check_args(args_)

        # remove leading '-' in option names
        misc = dict((k.lstrip("-"), v.strip("'"))
                    for k, v in zip(misc[::2], misc[1::2]))

        return args_, misc

    def _check_args(self, args_):
        """Checks command line arguments (e.g., whether all given files exist)
        Parameters
        ----------
        args_: parsed arguments
            Parsed command line arguments
        Raises
        ------
        ValueError
            in case of missing files or wrong configurations
        """

        if not os.path.isfile(args_.scenario_file):
            raise ValueError("Not found: %s" % (args_.scenario_file))



import os

import numpy as np
import pandas as pd
from bokeh.io import output_file, save, show
from bokeh.models import BasicTickFormatter, BasicTicker, ColorBar, LinearColorMapper, ColumnDataSource, HoverTool
from bokeh.palettes import Inferno256, Category20
from bokeh.plotting import figure
from bokeh.transform import transform


def bokeh_heatmap_num(data, x_label, y_label, x_log, y_log, grid_column='zz'):
    """
    Plot heatmap for numerical values (resolution predefined by dataframe)
    Grid-values are the values being filled in the headt-map rectangles

    Parameters
    ----------
    data: pandas.Dataframe
        dataframe
    x_label, y_label: str
        labels for axes
    x_log, y_log: bool
        whether to plot axes in (natural) log
    grid_column: str
        name of column with grid data in dataframe
    """
    # Reshape into 1D array
    df = pd.DataFrame(data.stack(), columns=[grid_column]).reset_index()
    source = ColumnDataSource(df)
    mapper = LinearColorMapper(palette=Inferno256, low=df[grid_column].min(), high=df[grid_column].max())
    x_range = (data.index[0], data.index[-1])
    y_range = (data.columns[0], data.columns[-1])
    return _bokeh_heatmap(source, mapper, x_label, y_label, x_range, y_range, x_log=x_log, y_log=y_log)

def bokeh_heatmap_cat(data, x_label, y_label, grid_column='zz'):
    """
    Plot heatmap for categorical values (resolution defined by number of choices)
    Grid-values are the values being filled in the headt-map rectangles

    Parameters
    ----------
    data: pandas.Dataframe
        dataframe
    x_label, y_label: str
        labels for axes
    grid_column: str
        name of column with grid data in dataframe
    """
    # Reshape into 1D array
    df = pd.DataFrame(data.stack(), columns=[grid_column]).reset_index()
    source = ColumnDataSource(df)
    mapper = LinearColorMapper(palette=Inferno256, low=df[grid_column].min(), high=df[grid_column].max())
    x_range = [str(c) for c in data.index]
    y_range = [str(c) for c in reversed(data.columns)]
    return _bokeh_heatmap(source, mapper, x_label, y_label, x_range, y_range)

def _bokeh_heatmap(source, colormapper, x_label, y_label, x_range, y_range, x_log=False, y_log=False):
    p = figure(x_range=x_range, y_range=y_range,
               x_axis_type="log" if x_log else "linear",
               y_axis_type="log" if y_log else "linear",
               toolbar_location=None, tools="")
    p.rect(x=x_label, y=y_label, width=1, height=1, source=source,
           fill_color=transform('zz', colormapper),
           line_color=None)
    color_bar = ColorBar(color_mapper=colormapper, location=(0, 0),
                         ticker=BasicTicker(desired_num_ticks=20),
                         formatter=BasicTickFormatter(use_scientific=False))
    p.add_layout(color_bar, 'right')
    p.xaxis.axis_label = x_label
    p.yaxis.axis_label = y_label
    return p

def bokeh_boxplot(labels, mean, std, x_label, y_label, runtime, inc_indices):
    """
    Plot a boxplot in bokeh

    Parameters
    ----------
    labels: List[str]
        labels for the x-axis (for the different boxes)
    mean: List[float]
        means, same length as labels
    std: List[float]
        stds, same length as labels
    x_label, y_label: str
        axis-labels
    runtime: bool
        whether runtime is analyzed, if so will limit y-axis to 0 for std
    inc_indices: List[int]
        list of indices in values that are to be marked as incumbents

    Returns
    -------
    plot: bokeh.plotting.figure
        the bokeh figure
    """
    p = figure(x_range=labels, height=350, width=800)
    upper = mean + std
    lower = mean - std
    if runtime:
        lower = [0 if l < 0 else l for l in lower]
    p.vbar(labels, 0.7, mean, upper, fill_color="#E08E79", line_color="black")
    p.vbar(labels, 0.7, mean, lower, fill_color="#3B8686", line_color="black")
    if len(inc_indices) > 0:
        p.circle(list([labels[idx] for idx in inc_indices]), list([float(mean[idx]) for idx in inc_indices]),
                 name='incumbent', color='black')

    p.xaxis.axis_label = x_label
    p.yaxis.axis_label = y_label
    p.yaxis.formatter = BasicTickFormatter(use_scientific=False)

    return p

def bokeh_multiline(data, x_limits, y_limits, x_column, line_columns, y_label, z_label):
    """
    Bokeh multiline plot

    Parameters
    ----------
    data: pandas.DataFrame
        data to be plotted
    x_limits, y_limits: Tuple(float, float)
        limits for x and y
    x_column: str
        name of the column in dataframe to be plotted on x-axis
    line_columns: List[str]
        names of the columns to be plotted as lines
    y_label: str
        label for y-axis
    z_label: str
        name of the categorical parameter (for tooltips)

    Returns
    -------
    plot: bokeh.plotting.figure
        the bokeh figure
    """
    ys = [data[lc] for lc in line_columns]
    xs = [data[x_column] for _ in range((len(ys)))]
    palette = Category20[len(xs)] if len(xs) > 2 else ['blue', 'red'] if len(xs) == 2 else ['blue']
    source = ColumnDataSource(dict(
        xs=xs,
        ys=ys,
        color=palette,
        label=line_columns,
    ))

    p = figure(x_range=x_limits,
               y_range=y_limits,
               toolbar_location=None, tools="")

    p.multi_line(xs="xs", ys="ys", color="color",
                 legend="label",
                 source=source)
    p.add_tools(HoverTool(show_arrow=False, line_policy='next', tooltips=[
        (z_label, '@label')
    ]))

    p.xaxis.axis_label = x_column
    p.yaxis.axis_label = y_label

    return p

def bokeh_line_uncertainty(values, mean, std, x_log, x_label, y_label, inc_indices):
    """
    Return a bokeh plot with a simple line plot and uncertainty marked as filled area above and below plot.

    Parameters
    ----------
    values: List
        values for x-axis (same length as mean and std)
    mean: List[float]
        means (same length as values and std)
    std: List[float]
        uncertainty to be plotted around the mean (same length as values and mean)
    x_log: bool
        plot x-axis in log
    x_label, y_label: str
        axis-labels
    inc_indices: List[int]
        list of indices in values that are to be marked as incumbents

    Returns
    -------
    plot: bokeh.plotting.figure
        the bokeh figure
    """
    lower_curve = mean - std
    upper_curve = mean + std

    p = figure(x_range=(values[0], values[-1]), x_axis_type="log" if x_log else "linear", height=350, width=800)
    p.line(values, mean)
    band_x = np.append(values, values[::-1])
    band_y = np.append(lower_curve, upper_curve[::-1])
    p.patch(band_x, band_y, color='#7570B3', fill_alpha=0.2)

    if len(inc_indices) > 0:
        p.circle(list([values[idx] for idx in inc_indices]), list([float(mean[idx]) for idx in inc_indices]),
                 name='incumbent', color='black')

    p.xaxis.axis_label = x_label
    p.yaxis.axis_label = y_label
    p.yaxis.formatter = BasicTickFormatter(use_scientific=False)

    return p

def save_and_show(plot_name, show_plot, layout):
    """
    Save `layout` in `plot_name` (which has to be the complete path including the filename, '.html' will be appended if
    necessary) and show plot in default browser if `show_plot` is `True`.
    """
    if plot_name:
        outdir = os.path.dirname(plot_name)
        if not os.path.exists(outdir):
            os.mkdir(outdir)
        if not plot_name.endswith('.html'):
            plot_name = plot_name + '.html'
        output_file(plot_name)
        save(layout)
    if show_plot:
        show(layout)

def shorten_unique(names, keep_first=4, keep_last=4):
    """
    Shorten strings, inserting '(...)', while keeping them unique.

    Parameters
    ----------
    names: List[str]
        list of strings to be shortened
    keep_first: int
        always keep the first N letters
    keep_last: int
        always keep the last N letters

    Returns
    -------
    shortened_names: List[str]
        list with shortened strings
    """
    short, cut_chars, longest_str = [], 0, max([len(p) for p in names])
    while len(set(short)) != len(set(names)) and cut_chars <= longest_str:
        short = [p[:keep_first + cut_chars] + '(...)' + p[-keep_last:]
                 if len(p) > sum([keep_first, keep_last, cut_chars])
                 else p for p in names]
        cut_chars += 1
    return short



import copy
from collections import OrderedDict

import matplotlib as mpl
import numpy as np
from bokeh.models import ColumnDataSource, CDSView, IndexFilter, Span, BasicTickFormatter, Row
from bokeh.plotting import figure

from pimp.utils.bokeh_helpers import save_and_show, shorten_unique

mpl.use('Agg')
from matplotlib import pyplot as plt

from pimp.configspace import AndConjunction, Configuration, OrConjunction, impute_inactive_values
from pimp.evaluator.base_evaluator import AbstractEvaluator

__author__ = "Andre Biedenkapp"
__copyright__ = "Copyright 2016, ML4AAD"
__license__ = "3-clause BSD"
__maintainer__ = "Andre Biedenkapp"
__email__ = "biedenka@cs.uni-freiburg.de"


class Ablation(AbstractEvaluator):

    """
    Implementation of Ablation via surrogates
    """

    def __init__(self, scenario, cs, model, to_evaluate: int, rng, incumbent=None, **kwargs):
        super().__init__(scenario, cs, model, to_evaluate, rng, **kwargs)
        self.name = 'Ablation'
        self.logger = 'pimp.' + self.name

        self.target = incumbent
        self.source = self.cs.get_default_configuration()
        self.target_active = {}
        self.source_active = {}
        self.delta, self.active = self._diff_in_source_and_target()
        self.default = self.cs.get_default_configuration()
        self._determine_combined_flipps()
        self.inactive = []

        params = self.cs.get_hyperparameters()
        self.n_params = len(params)
        self.n_feats = len(self.types) - len(params)
        self.insts = copy.deepcopy(self.scenario.train_insts)
        if len(self.scenario.test_insts) > 1:
            self.insts.extend(self.scenario.test_insts)
        self.predicted_parameter_performances = OrderedDict()
        self.predicted_parameter_variances = OrderedDict()

########################################################################################################################
    # HANDLING FORBIDDENS # HANDLING FORBIDDENS # HANDLING FORBIDDENS # HANDLING FORBIDDENS # HANDLING FORBIDDENS
########################################################################################################################
    def determine_forbidden(self):
        """
        Method to determine forbidden clauses, and saves them in a simple to check format.
        return: list of lists [[pname, pvalue, pname2, pvalue2]]
        """
        forbidden_clauses = self.cs.forbidden_clauses
        forbidden_descendants = map(lambda x: x.get_descendant_literal_clauses(), forbidden_clauses)
        forbidden_names_value_paris = list()
        for forbidden_literals in forbidden_descendants:
            elem = []
            for literal in forbidden_literals:
                elem.append(literal.hyperparameter.name)
                elem.append(literal.value)
            forbidden_names_value_paris.append(elem)
        return forbidden_names_value_paris

    def check_not_forbidden(self, forbidden_name_value_pairs, modifiable_config):
        """
        Helper method to determine if a current configuration dictionary is forbidden or not
        forbidden_name_value_pairs: list of lists of forbidden parameter settings
        modifiable_config: dict param_name -> param_value
        return: boolean. True if the current configuration is not forbidden, False otherwise.
        """
        not_forbidden = True
        for forbidden_clause in forbidden_name_value_pairs:
            sum_forbidden = 0
            for key in modifiable_config:
                if key in forbidden_clause:
                    at_ = forbidden_clause.index(key) + 1
                    if modifiable_config[key] == forbidden_clause[at_]:
                        sum_forbidden += 2
            not_forbidden = not_forbidden and not (sum_forbidden == len(forbidden_clause))
            if not not_forbidden:
                return False
        return True

########################################################################################################################
    # HANDLING CONDITIONALITIES # HANDLING CONDITIONALITIES # HANDLING CONDITIONALITIES # HANDLING CONDITIONALITIES
########################################################################################################################
    def _helper(self, modified_delta):
        to_remove = []
        indices_to_check = [i for i in range(len(modified_delta))]
        for at, idx in enumerate(indices_to_check):
            parameters = modified_delta[idx]
            for p in parameters:
                children = self.cs.get_children_of(p)
                for child in children:
                    if child.name not in modified_delta[idx]:
                        for condition in self.cs.get_parent_conditions_of(child):
                            if type(condition) in [AndConjunction, OrConjunction]:  # Case where parents of parents are
                                                                                    # not set get checked here
                                skip_this = False
                                for component in condition.components:
                                    if self.source.get(component.parent.name) is None:
                                        if [child.name] in modified_delta:
                                            if modified_delta.index([child.name]) not in to_remove:
                                                to_remove.append(modified_delta.index([child.name]))
                                        for tmp_idx, d in enumerate(modified_delta):
                                            if component.parent.name in d and child.name not in d:
                                                modified_delta[tmp_idx].append(child.name)
                                                indices_to_check.append(tmp_idx)
                                        skip_this = True
                                    if self.target.get(component.parent.name) is None:
                                        if [child.name] in modified_delta:
                                            if modified_delta.index([child.name]) not in to_remove:
                                                to_remove.append(modified_delta.index([child.name]))
                                        for tmp_idx, d in enumerate(modified_delta):
                                            if component.parent.name in d and child.name not in d:
                                                modified_delta[tmp_idx].append(child.name)
                                                indices_to_check.append(tmp_idx)
                                        skip_this = True
                                if skip_this:
                                    continue
                            try:
                                source_conditions_met = condition.evaluate(self.source)
                            except ValueError:
                                source_conditions_met = False
                            try:
                                target_conditions_met = condition.evaluate(self.target)
                            except ValueError:
                                target_conditions_met = False
                            if target_conditions_met and not source_conditions_met:
                                if child.name not in modified_delta[idx]:
                                    modified_delta[idx].append(child.name)  # Now at idx delta has two combined entries
                                    indices_to_check.append(idx)
                                try:
                                    tmp_idx = modified_delta.index([child.name])
                                    if [child.name] in modified_delta and tmp_idx not in to_remove:
                                        to_remove.append(modified_delta.index([child.name]))
                                except ValueError:
                                    pass
        return to_remove, modified_delta

    def _determine_combined_flipps(self):
        """
        Method to determine parameters that have to be jointly flipped with their parents.
        Uses the methods provided by Config space to easily check conditions
        """
        to_remove, self.delta = self._helper(self.delta)
        to_remove = sorted(to_remove, reverse=True)  # reverse sort necessary to not delete the wrong items
        for idx in to_remove:
            self.delta.pop(idx)
        to_remove = []
        single_remove = []
        for i in range(len(self.delta) - 1):
            flip_1 = self.delta[i]
            for j in range(i + 1, len(self.delta)):
                flip_2 = self.delta[j]
                for idx, entry in enumerate(flip_2):
                    if entry in flip_1:
                        if idx == 0:
                            to_remove.append(j)
                            self.logger.info('Removing %s' % str(self.delta[j]))
                        else:
                            single_remove.append((j, idx))
        single_remove = sorted(single_remove, reverse=True)
        for tuple_ in single_remove:
            del self.delta[tuple_[0]][tuple_[1]]
        to_remove = sorted(to_remove, reverse=True)  # reverse sort necessary to not delete the wrong items
        for idx in to_remove:
            self.delta.pop(idx)

    def _check_child_conditions(self, _dict, children):
        dict_ = {}
        for child in children:
            all_child_conditions_fulfilled = True
            for condition in self.cs.get_parent_conditions_of(child):
                try:
                    all_child_conditions_fulfilled = all_child_conditions_fulfilled and condition.evaluate(_dict)
                except ValueError:
                    all_child_conditions_fulfilled = False  # Parent not set!
            dict_[child.name] = all_child_conditions_fulfilled
        return dict_

    def _check_children(self, modded_dict, params, delete=False):
        """
        Method that checks if children are set to inactive during the current round due to condidionalities.

        Parameters
        ----------
        modded_dict: dictionary
            Dictionary with parameter_name -> parameter_value
        params: list
            list of parameters to check the child activity for
        delete: boolean
            allows for deletion of inactive parameters from the ablation_delta_list

        Returns
        -------
        modded_dict, where inactive parameters have the value None
        """
        for param in params:
            children = self.cs.get_children_of(param)
            if children:
                child_conditions = self._check_child_conditions(modded_dict, children)
                for child in child_conditions:
                    if not child_conditions[child]:
                        modded_dict[child] = None
                        self.logger.debug('Deactivated child %s found this round' % child)
                        modded_dict = self._set_child_of_child_to_none(child, modded_dict)
                        if delete and [child] in self.delta:
                            for item in self.delta:
                                self.logger.debug('%s, %s' % (str([child]), str(item)))
                            self.logger.critical('Removing deactivated parameter %s' % child)
                            self.delta.pop(self.delta.index([child]))
                    else:
                        if child not in self.target_active and child not in self.source_active:
                            modded_dict[child] = None
                        elif child in self.target_active:
                            if not self.target_active[child]:
                                if child in self.source_active:
                                    modded_dict[child] = self.source[child]
                                else:
                                    modded_dict[child] = None
                            elif self.target_active[child]:
                                modded_dict[child] = self.target[child]
                            else:
                                modded_dict[child] = self.cs.get_hyperparameter(child).default_value
                        modded_dict = self._check_children(modded_dict, [child])
        return modded_dict

    def _set_child_of_child_to_none(self, child, modded_dict):
        children_of_child = self.cs.get_children_of(child)
        if children_of_child:
            modded_dict[child] = self.target[child]
            try:
                conditions = self._check_child_conditions(modded_dict, children_of_child)
            except ValueError:
                modded_dict[child] = None
                return modded_dict
            modded_dict[child] = None
            for c in conditions:
                if not conditions[c] and c in list(modded_dict.keys()):
                    modded_dict[c] = None
                    self.logger.debug('Deactivated sub-child %s found this round' % c)
                    modded_dict = self._set_child_of_child_to_none(c, modded_dict)
        return modded_dict

    def _rm_inactive(self, candidates, modifiable_config_dict, prev_modifiable_config_dict, removed=[]):
        for candidate in candidates:
            parent_conditions = self.cs.get_parent_conditions_of(candidate)
            passing = True
            for condition in parent_conditions:
                try:
                    passing = condition.evaluate(modifiable_config_dict) and passing
                except ValueError:
                    passing = False
            if not passing:
                if candidate in modifiable_config_dict:
                    try:
                        modifiable_config_dict[candidate] = prev_modifiable_config_dict[candidate]
                    except KeyError:
                        del modifiable_config_dict[candidate]
                    removed.append(candidate)
                modifiable_config_dict, removed = self._rm_inactive(
                    list(map(lambda x: x.name, self.cs.get_children_of(candidate))), modifiable_config_dict,
                    prev_modifiable_config_dict, removed)
        return modifiable_config_dict, removed

########################################################################################################################
    # MAIN METHOD # MAIN METHOD # MAIN METHOD # MAIN METHOD # MAIN METHOD # MAIN METHOD # MAIN METHOD # MAIN METHOD
########################################################################################################################
    def run(self) -> OrderedDict:
        """
        Main function.
        Returns
        -------
        evaluated_parameter_importance:OrderedDict
            Parameter -> importance. The order is important as smaller indices indicate higher importance
        """
        # Minor setup
        modifiable_config_dict = copy.deepcopy(self.source.get_dictionary())
        prev_modifiable_config_dict = copy.deepcopy(self.source.get_dictionary())
        modified_so_far = []
        start_delta = len(self.delta)
        best_performance = -1

        # Predict source and target performance to later use it to predict the %improvement a parameter causes
        source_mean, source_var = self._predict_over_instance_set(impute_inactive_values(self.source))
        prev_performance = source_mean
        target_mean, target_var = self._predict_over_instance_set(impute_inactive_values(self.target))
        improvement = prev_performance - target_mean
        self.predicted_parameter_performances['-source-'] = source_mean.flatten()[0]
        self.predicted_parameter_variances['-source-'] = source_var.flatten()[0]
        self.evaluated_parameter_importance['-source-'] = 0

        forbidden_name_value_pairs = self.determine_forbidden()
        length_ = len(self.delta) - min(len(self.delta), self.to_evaluate)
        self.logger.info('Difference in source and target: %d' % len(self.delta))

        while len(self.delta) > length_:  # Main loop. While parameters still left ...
            modifiable_config_dict = copy.deepcopy(prev_modifiable_config_dict)
            self.logger.debug('Round %d of %d:' % (start_delta - len(self.delta) + 1, min(start_delta,
                                                                                          self.to_evaluate)))
            for param_tuple in modified_so_far:  # necessary due to combined flips
                for parameter in param_tuple:
                    modifiable_config_dict[parameter] = self.target[parameter]
            prev_modifiable_config_dict = copy.deepcopy(modifiable_config_dict)

            round_performances = []
            round_variances = []
            for candidate_tuple in self.delta:
                self.logger.debug('candidate(s): %s' % str(candidate_tuple))

                for candidate in candidate_tuple:
                    self.logger.debug(' {:<25s} -> {:^10s}'.format(str(candidate), str(self.target[candidate])))
                    modifiable_config_dict[candidate] = self.target[candidate]

                modifiable_config_dict = self._check_children(modifiable_config_dict, candidate_tuple)

                # Check if current config is allowed
                not_forbidden = self.check_not_forbidden(forbidden_name_value_pairs, modifiable_config_dict)
                if not not_forbidden:  # othwerise skipp it
                    self.logger.critical('FOUND FORBIDDEN!!!!! SKIPPING!!!')
                    continue
                try:
                    modifiable_config = Configuration(self.cs, modifiable_config_dict)
                except ValueError:
                    modifiable_config_dict, rmed = self._rm_inactive(candidate_tuple[1:], modifiable_config, prev_modifiable_config_dict)
                    modifiable_config = Configuration(self.cs, modifiable_config_dict)

                mean, var = self._predict_over_instance_set(impute_inactive_values(modifiable_config))  # ... predict their performance
                self.logger.debug('%s: %.6f' % (candidate_tuple, mean[0]))
                round_performances.append(mean)
                round_variances.append(var)
                modifiable_config_dict = copy.deepcopy(prev_modifiable_config_dict)

            best_idx = np.argmin(round_performances)
            assert 0 <= best_idx < len(round_performances), 'No improving parameter found!'
            best_performance = round_performances[best_idx]  # greedy choice of parameter to fix
            best_variance = round_variances[best_idx]
            improvement_in_percentage = (prev_performance - best_performance) / improvement
            prev_performance = best_performance
            modified_so_far.append(self.delta[best_idx])
            self.logger.info('Round %2d winner(s): (%s, %.4f)' % (start_delta - len(self.delta) + 1,
                                                                  str(self.delta[best_idx]),
                                                                  improvement_in_percentage * 100))
            param_str = '; '.join(self.delta[best_idx])
            self.evaluated_parameter_importance[param_str] = improvement_in_percentage.flatten()[0]
            self.predicted_parameter_performances[param_str] = best_performance.flatten()[0]
            self.predicted_parameter_variances[param_str] = best_variance.flatten()[0]

            for winning_param in self.delta[best_idx]:  # Delete parameters that were set to inactive by the last
                                                        # best parameter

                prev_modifiable_config_dict[winning_param] = self.target[winning_param]
            self._check_children(prev_modifiable_config_dict, self.delta[best_idx], delete=True)
            self.delta.pop(best_idx)  # don't forget to remove already tested parameters

        self.predicted_parameter_performances['-target-'] = target_mean.flatten()[0]
        self.predicted_parameter_variances['-target-'] = target_var.flatten()[0]
        self.evaluated_parameter_importance['-target-'] = 0
        # sum_ = 0  # Small check that sum is 1
        # for key in self.evaluated_parameter_importance:
        #     print(key, self.evaluated_parameter_importance[key])
        #     sum_ += self.evaluated_parameter_importance[key]
        # print(sum_)
        all_res = {'perf': self.predicted_parameter_performances, 'var': self.predicted_parameter_variances,
                'imp': self.evaluated_parameter_importance, 'order': list(self.evaluated_parameter_importance.keys())}
        return all_res

########################################################################################################################
# HELPER METHODS # HELPER METHODS # HELPER METHODS # HELPER METHODS # HELPER METHODS # HELPER METHODS # HELPER METHODS
########################################################################################################################
    def _predict_over_instance_set(self, config):
        """
        Small wrapper to predict marginalized over instances
        Parameter
        ---------
        config:Configuration
            The configuration of wich the performance across the whole instance set is to be estimated
        Returns
        -------
        mean
            the mean performance over the instance set
        var
            the variance over the instance set. If logged values are used, the variance might not be able to be used
        """
        mean, var = self.model.predict_marginalized_over_instances(np.array([config.get_array()]))
        return mean, var

    def _diff_in_source_and_target(self):
        """
        Helper Method to determine which parameters might lie on an ablation path
        Return
        ------
        delta:list
            List of parameters that are modified from the source to the target
        """
        delta = []
        active = {}
        for parameter in self.cs.get_hyperparameters():
            parameter = parameter.name
            tmp = ' not'
            if self.source[parameter] != self.target[parameter] and self.target[parameter] is not None:
                tmp = ''
                delta.append([parameter])
            self.logger.debug('%s was%s modified from source to target (%s, %s) [s, t]' % (parameter, tmp,
                                                                                           self.source[parameter],
                                                                                           self.target[parameter]))
            self.target_active[parameter] = True if self.target[parameter] is not None else False
            active[parameter] = True if self.source[parameter] is not None else False
        self.source_active = copy.deepcopy(active)
        return delta, active

########################################################################################################################
# PLOTTING # PLOTTING # PLOTTING # PLOTTING # PLOTTING # PLOTTING # PLOTTING # PLOTTING # PLOTTING # PLOTTING # PLOTTING
########################################################################################################################
    def plot_result(self, name=None, show=True):
        self.plot_predicted_percentage(plot_name=name+'percentage.png', show=show)
        self.plot_predicted_performance(plot_name=name+'performance.png', show=show)
        self.logger.info('Saved plots as %s[percentage|performance].png' % name)
        plt.close('all')


    def plot_bokeh(self, plot_name=None, show_plot=False, predicted_percentage=True, predicted_performance=True):
        """
        Plot ablation results in an interactive bokeh-plot.

        Parameters
        ----------
        plot_name: str
            path where to store the plot, None to not save it
        show_plot: bool
            whether or not to open plot in standard browser
        predicted_percentage: bool
            include a barchart of individual parameter contributions of the improvement from source to target
        predicted_performance: bool
            include a plot of the ablation path using the predicted performances of parameter flips

        Returns
        -------
        layout: bokeh.models.Row
            bokeh plot (can be used in notebook or comparted with components)
        """

        # Get all relevant data-points
        p_names = list(self.evaluated_parameter_importance.keys())
        p_names_short = shorten_unique(p_names)
        p_importance = [100 * self.evaluated_parameter_importance[p] for p in p_names]
        p_performance = [self.predicted_parameter_performances[p] for p in p_names]
        p_variance = [self.predicted_parameter_variances[p] for p in p_names]

        # We don't always want to plot everything, reduce plotted parameters by taking out the least important ones
        max_to_plot = min(len(p_names), self.MAX_PARAMS_TO_PLOT)
        plot_indices = sorted(range(1, len(p_importance) - 1), key=lambda x: p_importance[x], reverse=True)[:max_to_plot]
        plot_indices = [0] + plot_indices + [len(p_importance) - 1]  # Always keep "source" and "target"

        # Customizing plot-style
        bar_width = 25
        tooltips = [("Parameter", "@parameter_names"),
                    ("Importance", "@parameter_importance"),
                    ("Predicted Performance", "@parameter_performance")]

        # Create ColumnDataSource for both plots
        source = ColumnDataSource(data=dict(parameter_names_short=p_names_short,
                                            parameter_names=p_names,
                                            parameter_importance=p_importance,
                                            parameter_performance=p_performance,
                                            parameter_variance=p_variance,
                                            ))

        plots = []
        if predicted_percentage:
            # Don't plot source and target in this plot, use IndexFilter
            view_predicted_percentage = CDSView(source=source, filters=[IndexFilter(plot_indices[1:-1])])
            p = figure(x_range=[p_names_short[idx] for idx in plot_indices[1:-1]],
                       plot_height=350, plot_width=100+max_to_plot*bar_width,
                       toolbar_location=None, tools="hover", tooltips=tooltips)

            p.vbar(x='parameter_names_short', top='parameter_importance', width=0.9,
                   source=source, view=view_predicted_percentage)
            for value in [self.IMPORTANCE_THRESHOLD, -self.IMPORTANCE_THRESHOLD]:
                p.add_layout(Span(location=value*100, dimension='width', line_color='red', line_dash='dashed'))

            p.yaxis.axis_label = "improvement [%]"
            plots.append(p)

        if predicted_performance:
            view_predicted_performance = CDSView(source=source, filters=[IndexFilter(plot_indices)])
            p = figure(x_range=[p_names_short[idx] for idx in plot_indices],
                       plot_height=350, plot_width=100 + max_to_plot * bar_width,
                       toolbar_location=None, tools="hover", tooltips=tooltips)

            p.line(x='parameter_names_short', y='parameter_performance', source=source, view=view_predicted_performance)
            band_x = np.append(list(range(len(plot_indices))), list(range(len(plot_indices)))[::-1])
            band_y = np.append([max(p-np.sqrt(v), 0) if self.scenario.run_obj == "runtime" else p-np.sqrt(v)
                                for p, v in [(p_performance[idx], p_variance[idx]) for idx in plot_indices]],
                               [p+np.sqrt(v) for p, v in [(p_performance[idx], p_variance[idx])
                                                          for idx in plot_indices][::-1]])
            p.patch(band_x, band_y, color='#7570B3', fill_alpha=0.2)

            p.yaxis.axis_label = "est. runtime [sec]" if self.scenario.run_obj == 'runtime' else 'est. cost'
            plots.append(p)

        # Common styling:
        for p in plots:
            p.xaxis.major_label_orientation = 1.3
            p.xaxis.major_label_text_font_size = "14pt"
            p.yaxis.formatter = BasicTickFormatter(use_scientific=False)

        layout = Row(*plots)

        # Save and show...
        save_and_show(plot_name, show_plot, layout)

        return layout

    def plot_predicted_percentage(self, plot_name=None, show=True):
        """
        Method to plot a barchart of individual parameter contributions of the improvement from source to target
        """
        p_names = np.array(list(self.evaluated_parameter_importance.keys())[1:-1])
        p_names_short = np.array(shorten_unique(p_names))
        p_importance = np.array([100 * self.evaluated_parameter_importance[p] for p in p_names])
        max_to_plot = min(len(p_names), self.MAX_PARAMS_TO_PLOT)

        fig = plt.figure()
        ax1 = fig.add_subplot(111)
        plt.subplots_adjust(bottom=0.25, top=0.9, left=0.05, right=.95)

        ax1.bar(list(range(len(p_names))), p_importance, color=self.area_color)

        ax1.set_xticks(np.arange(len(p_names)))
        ax1.set_xlim(-0.5, max_to_plot - .25)
        ax1.set_ylim(min(p_importance) - max(2, min(p_importance)*.1),
                     max(p_importance) + min(2, max(p_importance)*.1))

        ax1.set_ylabel('improvement [%]', zorder=81, **self.LABEL_FONT)
        ax1.plot(list(range(-1, len(p_names_short) + 1)),
                 [self.IMPORTANCE_THRESHOLD*100 for _ in range(len(p_names_short) + 2)], c='r', linestyle='--')
        ax1.plot(list(range(-1, len(p_names_short) + 1)),
                 [-self.IMPORTANCE_THRESHOLD*100 for _ in range(len(p_names_short) + 2)], c='r', linestyle='--')
        ax1.set_xticklabels(p_names_short, rotation=25, ha='right', **self.AXIS_FONT)

        for idx, t in enumerate(ax1.xaxis.get_ticklabels()):
            color_ = (0.45, 0.45, 0.45)
            if self.evaluated_parameter_importance[p_names[idx]] > self.IMPORTANCE_THRESHOLD:
                color_ = (0., 0., 0.)
            t.set_color(color_)

        ax1.xaxis.grid(True)
        ax1.yaxis.grid(True)
        try:
            plt.tight_layout()
        except ValueError:
            pass

        if plot_name is not None:
            fig.savefig(plot_name)
            if show:
                plt.show()
        else:
            plt.show()

    def plot_predicted_performance(self, plot_name=None, show=True):
        """
        Method to plot the ablation path using the predicted performances of parameter flips
        """

        fig = plt.figure()
        ax1 = fig.add_subplot(111)
        plt.subplots_adjust(bottom=0.25, top=0.9, left=0.05, right=.95)
        y_label = self.scenario.run_obj if self.scenario.run_obj != 'quality' else 'cost'

        path = list(self.predicted_parameter_performances.keys())
        path = np.array(shorten_unique(path))
        max_to_plot = min(len(path), self.MAX_PARAMS_TO_PLOT)
        performances = list(self.predicted_parameter_performances.values())
        performances = np.array(performances).reshape((-1, 1))
        variances = list(self.predicted_parameter_variances.values())
        variances = np.array(variances).reshape((-1, 1))

        if max_to_plot == self.MAX_PARAMS_TO_PLOT:
            performances[self.MAX_PARAMS_TO_PLOT - 1] = performances[-1]
            variances[self.MAX_PARAMS_TO_PLOT - 1] = variances[-1]
            path[self.MAX_PARAMS_TO_PLOT - 1] = path[-1]

        ax1.plot(list(range(len(performances))), performances,
                 label='predicted %s' % y_label,
                 ls='-', zorder=80,
                 **self.LINE_FONT)

        upper = np.array(list(map(lambda x, y: x + np.sqrt(y), performances, variances))).flatten()
        if self.scenario.run_obj == "runtime":
            lower = np.array(list(map(lambda x, y: max(x - np.sqrt(y), np.array([0])), performances,
                                      variances))).squeeze()
        else:
            lower = np.array(list(map(lambda x, y: x - np.sqrt(y), performances,
                                      variances))).squeeze()
        ax1.fill_between(list(range(len(performances))), lower, upper, label='std', color=self.area_color)

        ax1.set_xticks(list(range(len(path))))
        ax1.set_xticklabels(path, rotation=25, ha='right', **self.AXIS_FONT)
        percentages = list(self.evaluated_parameter_importance.values())
        for idx, t in enumerate(ax1.xaxis.get_ticklabels()):
            color_ = (0.45, 0.45, 0.45)
            if percentages[idx] > self.IMPORTANCE_THRESHOLD:
                color_ = (0., 0., 0.)
            t.set_color(color_)

        ax1.set_xlim(0, max_to_plot - 1)
        if self.scenario.run_obj == "runtime":
            ax1.set_ylim(max(min(lower) - max(.1 * min(lower), 0.1), 0), max(upper) + .1 * max(upper))
        else:
            ax1.set_ylim(min(lower) - max(.1 * min(lower), 0.1), max(upper) + .1 * max(upper))

        ax1.legend()
        if self.scenario.run_obj == 'runtime':
            ax1.set_ylabel('runtime [sec]', zorder=81, **self.LABEL_FONT)
        else:
            ax1.set_ylabel('%s' % y_label, zorder=81, **self.LABEL_FONT)
        ax1.xaxis.grid(True)
        ax1.yaxis.grid(True)
        try:
            plt.tight_layout()
        except ValueError:
            pass

        if plot_name is not None:
            fig.savefig(plot_name)
            if show:
                plt.show()
        else:
            plt.show()



from pimp.epm.epar_x_rfwi import EPARrfi
from pimp.epm.unlogged_rfwi import Unloggedrfwi as Urfi

__author__ = "Andre Biedenkapp"
__copyright__ = "Copyright 2016, ML4AAD"
__license__ = "3-clause BSD"
__maintainer__ = "Andre Biedenkapp"
__email__ = "biedenka@cs.uni-freiburg.de"


class UnloggedEPARXrfi(EPARrfi, Urfi):

    def __init__(self, configspace, types, bounds, seed,
                 cutoff=0,
                 threshold=0, **kwargs):
        """
        TODO
        """
        Urfi.__init__(self, configspace=configspace, types=types, bounds=bounds, seed=seed, **kwargs)
        EPARrfi.__init__(self, configspace=configspace, types=types, bounds=bounds, seed=seed, cutoff=cutoff, threshold=threshold, **kwargs)

    def predict(self, X):
        """
        Method to override the predict method of RandomForestWithInstances.
        Thus it can be used in the marginalized over instances method of the RFWI class
        """
        return self._predict_EPAR(self._predict(X))



from ConfigSpace import ConfigurationSpace, Configuration
from ConfigSpace.read_and_write import pcs
from ConfigSpace.util import impute_inactive_values, get_random_neighbor,\
    get_one_exchange_neighbourhood
from ConfigSpace.c_util import change_hp_value, check_forbidden
from ConfigSpace.hyperparameters import CategoricalHyperparameter, FloatHyperparameter, IntegerHyperparameter
from ConfigSpace.conditions import AndConjunction, OrConjunction, InCondition, EqualsCondition
from ConfigSpace.exceptions import ForbiddenValueError
from ConfigSpace.hyperparameters import UniformFloatHyperparameter, UniformIntegerHyperparameter, Constant, \
    OrdinalHyperparameter, NumericalHyperparameter



import copy
import random
import time
from collections import OrderedDict

import numpy as np
import matplotlib as mpl
mpl.use('Agg')
from matplotlib import pyplot as plt
from sklearn.linear_model import LinearRegression
from sklearn.metrics import mean_squared_error

from pimp.evaluator.base_evaluator import AbstractEvaluator

__author__ = "Andre Biedenkapp"
__copyright__ = "Copyright 2016, ML4AAD"
__license__ = "3-clause BSD"
__maintainer__ = "Andre Biedenkapp"
__email__ = "biedenka@cs.uni-freiburg.de"


class InfluenceModel(AbstractEvaluator):

    """
    Implementation of Influence Models
    """

    def __init__(self, scenario, cs, model, to_evaluate: int, rng, margin: float=None, threshold: float=None, **kwargs):
        super().__init__(scenario, cs, model, to_evaluate, rng, **kwargs)
        self.name = 'InfluenceModel'
        self.logger = 'pimp.' + self.name
        self.model = LinearRegression()
        self.smac_model = model  # TODO Only use the SMAC model as container for X and y.
        self.all_X = copy.deepcopy(self.X)
        self.all_y = copy.deepcopy(self.y)

        # TODO better splitting
        self.testX = self.X[:int(len(self.X) * .1), :]
        self.X = self.X[int(len(self.X) * .1):, :]
        self.testy = self.y[:int(len(self.y) * .1)]
        self.y = self.y[int(len(self.y) * .1):]

        self.threshold = 0.000005 if threshold is None else threshold
        self.margin = 0.00000005 if margin is None else margin

    def _generate_random_trajectory(self, params, param_ids):
        """
        Simple method to shuffle around order in which parameters will be evaluated
        :param params:
        :param param_ids:
        :return:
        """
        tmp = list(range(len(params)))
        random.shuffle(tmp)
        return [params[i] for i in tmp], [param_ids[i] for i in tmp]

    def run(self) -> OrderedDict:
        """
        Implementation of the Stepwise feature selection as presented in
        "Performance-Influence Models for Highly Configurable Systems"
        :return: OrderedDict (param_name) -> (importance_value)
        """
        # Setup
        error = float('inf')
        lastError = float('inf')
        start_params = self.cs.get_hyperparameters()
        start_param_ids = list(range(len(start_params)))
        params = copy.deepcopy(start_params)
        param_ids = copy.deepcopy(start_param_ids)
        used = []  # Contains ids of which parametes will be used per round. Used for list-slicing
        # used.extend(range(len(params), len(self.smac_model.types)))  # we don't want to evaluate the feature importance

        # Beginning of the forward pass
        for num_evaluated in range(self.to_evaluate):
            best = None
            lastError = error
            params, param_ids = self._generate_random_trajectory(params, param_ids)

            counter = -1
            for idx, parameter in zip(param_ids, params):  # Along this trajectory sample the performance of parameters
                counter += 1
                self.logger.debug('Evaluating %s' % parameter.name)
                used.append(idx)  # similar to forward selection
                self.logger.debug('Used parameters: %s' % str(used))

                start = time.time()
                self.model.fit(self.X[:, used], self.y)  # refit the model every round
                round_error = mean_squared_error(self.testy, self.model.predict(self.testX[:, used]))
                self.logger.debug('Refitted LR (sec %.2f; error: %.4f)' % (time.time() - start, round_error))
                used.pop()
                if round_error < error:
                    error = round_error
                    best = parameter
                    best_idx = counter  # if we set it to idx we get the idx in start_params not which parameter to pop
                                        # from params and param_ids
            if error < self.threshold or (lastError - error) < self.margin:
                # If no imporvement is found this method automatically stops
                self.logger.debug('Stopping Forward pass')
                self.logger.debug('error < threshold: %s' % str(error < self.threshold))
                self.logger.debug('\delta error : %s' % str((lastError - error)))
                self.logger.debug('\delta error < margin: %s' % str((lastError - error) < self.margin))
                break
            else:
                if best is not None:
                    params.pop(best_idx)
                    used.append(param_ids.pop(best_idx))
                    self.logger.info('%s: %.4f (error)' % (best.name, error))
                    self.evaluated_parameter_importance[best.name] = self.model.coef_[-1]
                    best = None
                    best_idx = None
        remove = 0
        while remove >= 0:  # backward pass to determine if one parameter grew obsolete by adding others
            remove = -1
            for i in range(len(used)):
                tmp_used = used[:i] + used[i:]
                self.logger.debug('Used parameters: %s' % str(tmp_used))

                start = time.time()
                self.model.fit(self.X[:, used], self.y)  # refit the model every round
                round_error = mean_squared_error(self.testy, self.model.predict(self.testX[:, used]))
                # self.evaluated_parameter_importance[used[i]] = error
                self.logger.debug('Refitted LR (sec %.2f; error: %.4f)' % (time.time() - start, round_error))
                if round_error < error:  # if an obsolete parameter has been found.
                    error = round_error
                    remove = i  # remove it
                    del self.evaluated_parameter_importance[start_params[used[remove]].name]
                    self.logger.info('Removed %s' % start_params[used[remove]].name)
                    break
            if remove >= 0:
                used.pop(remove)
        self.model.fit(self.all_X[:, used], self.all_y)
        self.evaluated_parameter_importance = OrderedDict()
        coeffs = self.model.coef_.flatten()
        for idx, id in enumerate(used):
            self.evaluated_parameter_importance[start_params[id].name] = coeffs[idx]

        all_res = {'imp': self.evaluated_parameter_importance, 'order': list(self.evaluated_parameter_importance.keys())}
        return all_res

    def plot_result(self, name=None, show=True):
        # TODO find out what a sensible way of plotting this would be.
        # get sort index
        ids = [i[0] for i in sorted(enumerate(self.evaluated_parameter_importance.values()), key=lambda x: x[1])]

        fig, ax = plt.subplots()
        params = np.array(list(self.evaluated_parameter_importance.keys()))[ids]
        weights = np.array(list(self.evaluated_parameter_importance.values()))[ids]

        max_weight = max(weights) + abs(min(weights))
        tmp = np.arange(len(weights))
        bars = ax.bar(tmp, weights, color=self.area_color)

        for b, t in zip(enumerate(bars), ax.xaxis.get_ticklabels()):
            if (abs(weights[b[0]]) / max_weight) < 2*self.IMPORTANCE_THRESHOLD:
                b[1].set_color(self.unimportant_area_color)
                t.set_color((0.45, 0.45, 0.45))

        ax.set_ylabel('weights', **self.LABEL_FONT)
        ax.set_xticks(tmp)
        ax.set_xlim(-.5, len(tmp) - 0.5)
        ax.set_xticklabels(params, rotation=30, ha='right', **self.AXIS_FONT)
        ax.xaxis.grid(True)
        ax.yaxis.grid(True)

        plt.ylim((-1*(abs(min(weights)) + .1*abs(min(weights))), abs(max(weights)) + .1*abs(max(weights))))

        plt.tight_layout()
        if name is not None:
            fig.savefig(name)
            if show:
                plt.show()
            self.logger.info('Saved plot as %s.png' % name)
        else:
            plt.show()
        plt.close('all')



import abc
import logging
from collections import OrderedDict

from smac.epm.rf_with_instances import RandomForestWithInstances
from ConfigSpace.hyperparameters import UniformFloatHyperparameter

from pimp.configspace import ConfigurationSpace
from pimp.utils import Scenario

__author__ = "Andre Biedenkapp"
__copyright__ = "Copyright 2016, ML4AAD"
__license__ = "3-clause BSD"
__maintainer__ = "Andre Biedenkapp"
__email__ = "biedenka@cs.uni-freiburg.de"


class AbstractEvaluator(object):
    """
    Abstract implementation of Importance evaluator
    """
    def __init__(self, scenario: Scenario,
                 cs: ConfigurationSpace,
                 model: RandomForestWithInstances,
                 to_evaluate: int, rng,
                 verbose: bool=True,
                 **kwargs):
        self._logger = None
        self.scenario = scenario
        self.cs = cs
        self.model = model  # SMAC model
        self.rng = rng
        self.verbose = verbose

        if self.model is not None:
            if 'X' in kwargs and 'y' in kwargs:
                self._train_model(kwargs['X'], kwargs['y'], **kwargs)
            if 'features' in kwargs:
                self.features = kwargs['features']
            else:
                self.features = self.model.instance_features

            self.X = self.model.X
            self.y = self.model.y
            self.types = self.model.types
            self.bounds = self.model.bounds
        self._to_eval = to_evaluate
        if to_evaluate <= 0:
            self.to_evaluate = len(self.cs.get_hyperparameters())
        elif to_evaluate >= len(self.cs.get_hyperparameters()):
            self.to_evaluate = len(self.cs.get_hyperparameters())
        else:
            self.to_evaluate = to_evaluate  # num of parameters to evaluate

        self.evaluated_parameter_importance = OrderedDict()
        self.name = 'Base'

        self.IMPORTANCE_THRESHOLD = 0.05
        self.AXIS_FONT = {'family': 'monospace'}
        self.LABEL_FONT = {'family': 'sans-serif'}
        self.LINE_FONT = {'lw': 4,
                          'color': (0.125, 0.125, 0.125)}
        self.area_color = (0.25, 0.25, 0.45)
        self.unimportant_area_color = (0.125, 0.125, 0.225)
        self.MAX_PARAMS_TO_PLOT = 15

    @abc.abstractclassmethod
    def run(self) -> OrderedDict:
        raise NotImplementedError

    @abc.abstractclassmethod
    def plot_result(self, name=None):
        raise NotImplementedError

    def _train_model(self, X, y, **kwargs):
        self.model.train(X, y, **kwargs)

    def __str__(self):
        tmp = 'Parameter Importance Evaluation Method %s\n' % self.name
        tmp += '{:^15s}: {:<8s}\n'.format('Parameter', 'Value')
        for key in self.evaluated_parameter_importance:
            value = self.evaluated_parameter_importance[key]
            tmp += '{:>15s}: {:<3.4f}\n'.format(key, value)
        return tmp

    @property
    def logger(self):
        return self._logger

    @logger.setter
    def logger(self, value):
        self._logger = logging.getLogger(value)

    def _refit_model(self, types, bounds, X, y):
        """
        Easily allows for refitting of the model.
        Parameters
        ----------
        types: list
            SMAC EPM types
        X:ndarray
            X matrix
        y:ndarray
            corresponding y vector
        """
        # We need to fake config-space bypass imputation of inactive values in random forest implementation
        fake_cs = ConfigurationSpace(name="fake-cs-for-configurator-footprint")
        # We need to add fake hyperparameters
        fake_cs.add_hyperparameters([
            UniformFloatHyperparameter('fake-%s' % i, lower=0., upper=100000., default_value=0.,
                                       log=False) for i in range(len(types))
        ])

        self.model = RandomForestWithInstances(fake_cs, types, bounds, seed=12345, do_bootstrapping=True)
        self.model.rf_opts.compute_oob_error = True
        self.model.train(X, y)



import os
import shutil
import json
import glob
import itertools
from collections import OrderedDict as od
import numpy as np
from pandas import DataFrame, set_option as pd_option, concat as pd_concat
import pandas as pd


def jon_dirs(directory):
    if not isinstance(directory, list):
        directory = [directory]
    for d in directory:
        p_dir = os.getcwd()
        os.chdir(d)
        dir_ = sorted(os.listdir('.'))
        count = 0
        for entry in dir_:
            if '_all' in entry:
                count = 0
                continue
            if os.path.isdir(entry):
                tmp = entry.split('_')
                if '2017' in entry:
                    tmp = tmp[:tmp.index('2017') - 1]
                    tmp.extend(['rerun', 'all'])
                else:
                    tmp = tmp[:-1]
                    tmp.extend(['all'])
                base = '_'.join(tmp)
                print(base)

                if not os.path.exists(base):
                    os.mkdir(base)
                    count += 1
                if count > 0:
                    for file in os.listdir(entry):
                        shutil.move(os.path.join(entry, file), os.path.join(base, file))
                    os.removedirs(entry)
        os.chdir(p_dir)


def load_data_frames(directory, rerun=False, show=True):
    pd_option('max_colwidth', 120)
    dirs = glob.glob(os.path.join(directory, '*/'))

    data_dict = {}
    name_dict = {}
    keys = sorted(['fanova', 'fanova_cut', 'ablation', 'forward-selection', 'incneighbor'])
    for dir_ in dirs:
        if rerun and 'rerun' not in dir_:
            continue
        elif not rerun and 'rerun' in dir_:
            continue
        if show:
            print('Loading data from %s' % dir_)
        dn = dir_.split(os.path.sep)[-2]
        data_dict[dn] = {'fanova': None, 'fanova_cut': None, 'ablation': None, 'incneighbor': None}
        name_dict[dn] = {'fanova': None, 'fanova_cut': None, 'ablation': None, 'incneighbor': None}
        json_files = glob.glob(os.path.join(dir_, '*.json'))
        try:
            rm = json_files.index(os.path.join(dir_, 'pimp_args.json'))
            del json_files[rm]
        except ValueError:
            pass
        json_files = sorted(json_files)
        key_idx = 0
        jso_idx = 0
        while key_idx < len(keys) and jso_idx < len(json_files):
            if keys[key_idx] in ['forward-selection']:
                key_idx += 1
                continue
            cut = False
            if keys[key_idx] in json_files[jso_idx]:
                with open(json_files[jso_idx], 'r') as fh:
                    tmp = json.load(fh)
                if keys[key_idx] == 'fanova_cut':
                    keys[key_idx] = 'fanova'
                    cut = True
                importance = od([(k, tmp[keys[key_idx]]['imp'][k]) for k in tmp[keys[key_idx]]['order']])
                name = [k for k in tmp[keys[key_idx]]['order']]
                if keys[key_idx] == 'ablation':
                    if '-target-' in importance:
                        del importance['-target-']
                        del name[-1]
                    if '-source-' in importance:
                        del importance['-source-']
                        del name[0]
                if cut:
                    keys[key_idx] = 'fanova_cut'
                data_dict[dn][keys[key_idx]] = importance
                name_dict[dn][keys[key_idx]] = name
                key_idx += 1
                jso_idx += 1
            elif key_idx > jso_idx:
                jso_idx += 1
            else:
                key_idx += 1
    df = DataFrame.from_dict(data_dict, orient='index')
    dff = DataFrame.from_dict(name_dict, orient='index')
    cols = list(df.columns)
    cols[cols.index('incneighbor')] = 'local improvement analysis'
    df.columns = cols
    dff.columns = cols
    return df, dff


def _on_instance_sets(pd, m, index, store_in=None, algo_name=None, show=True, diagonal=False):
    first = sorted(index)
    if not diagonal:
        first = sorted(index[:-1])
    for idx, i in enumerate(first):
        a = pd[m][i]
        if a is None:
            continue
        a_keys = np.array(list(a.keys()))
        a_vals = np.array(list(a.values()))
        for at in range(len(a_keys)):
            if '[' in a_keys[at]:
                a_keys = a_keys[:at]
                a_vals = a_vals[:at]
                break
        a_indices_1 = np.array(np.nonzero(a_vals >= 0.01)).squeeze()
        a_indices_5 = np.array(np.nonzero(a_vals >= 0.05)).squeeze()
        a_keys_1 = a_keys[a_indices_1].reshape(1, -1).flatten()
        a_keys_5 = a_keys[a_indices_5].reshape(1, -1).flatten()
        second = sorted(index)
        if not diagonal:
            second = sorted(index[idx + 1:])
        for j in second:
            b = pd[m][j]
            if b is None:
                continue
            if algo_name:
                i = i.replace('_' + algo_name, '')
                i = i.replace(algo_name + '_', '')
                i = i.replace(algo_name, '')
                j = j.replace('_' + algo_name, '')
                j = j.replace(algo_name + '_', '')
                j = j.replace(algo_name, '')
            i = i.replace('_all', '')
            if '\\_' not in i:
                i = i.replace('_', '\\_')
            j = j.replace('_all', '')
            if '\\_' not in j:
                j = j.replace('_', '\\_')
            if show:
                print(i, j)
            b_keys = np.array(list(b.keys()))
            b_vals = np.array(list(b.values()))
            for at in range(len(b_keys)):
                if '[' in b_keys[at]:
                    b_keys = b_keys[:at]
                    b_vals = b_vals[:at]
                    break
            b_indices_1 = np.array(np.nonzero(b_vals >= 0.01)).squeeze()
            b_indices_5 = np.array(np.nonzero(b_vals >= 0.05)).squeeze()
            b_keys_1 = b_keys[b_indices_1].reshape(1, -1).flatten()
            b_keys_5 = b_keys[b_indices_5].reshape(1, -1).flatten()
            intersection_1 = list(set.intersection(set(a_keys_1), set(b_keys_1)))
            union_1 = list(set.union(set(a_keys_1), set(b_keys_1)))
            tmp = []
            for key in intersection_1:
                if ';' in key:
                    tmp.extend(key.split('; '))
                else:
                    tmp.append(key)
            intersection_1 = np.array(tmp).flatten()
            tmp = []
            for key in union_1:
                if ';' in key:
                    tmp.extend(key.split('; '))
                else:
                    tmp.append(key)
            union_1 = np.array(tmp).flatten()
            intersection_5 = list(set.intersection(set(a_keys_5), set(b_keys_5)))
            union_5 = list(set.union(set(a_keys_5), set(b_keys_5)))
            tmp = []
            for key in intersection_5:
                if ';' in key:
                    tmp.extend(key.split('; '))
                else:
                    tmp.append(key)
            intersection_5 = np.array(tmp).flatten()
            tmp = []
            for key in union_5:
                if ';' in key:
                    tmp.extend(key.split('; '))
                else:
                    tmp.append(key)
            union_5 = np.array(tmp).flatten()
            if show:
                print('          >=5%/>=1%')
                print('intersect: %3d/%3d' % (len(intersection_5), len(intersection_1)))
                print('    union: %3d/%3d' % (len(union_5), len(union_1)))
                print()
            if store_in is not None:
                if m not in store_in:
                    store_in[m] = {}
                store_in[m][' / '.join([i, j])] = [len(intersection_5), len(union_5)]
    return store_in


def _pairwise(pd, methods, index, store_in=None, algo_name=None, show=True):
    for i in sorted(index):
        skip = False
        intersection_set_1 = []
        intersection_set_5 = []
        intersection_set_f = []
        intersection_set_a = []
        for m in methods:
            entry = pd[m][i]
            if entry is None:
                skip = True
                break
            keys = np.array(list(entry.keys()))
            vals = np.array(list(entry.values()))
            for at in range(len(keys)):
                if '[' in keys[at]:
                    keys = keys[:at]
                    vals = vals[:at]
                    break
            indices_1 = np.array(np.nonzero(vals >= 0.01)).squeeze()
            indices_5 = np.array(np.nonzero(vals >= 0.05)).squeeze()
            keys_1 = keys[indices_1].reshape(1, -1).flatten()
            keys_1 = [x for y in list(map(lambda x: x.split('; '), keys_1)) for x in y]
            keys_5 = keys[indices_5].reshape(1, -1).flatten()
            keys_5 = [x for y in list(map(lambda x: x.split('; '), keys_5)) for x in y]
            intersection_set_1.append(set(keys_1))
            intersection_set_5.append(set(keys_5))
            if m == 'ablation':
                intersection_set_a.append(set(keys_1))
                intersection_set_f.append(set(keys_5))
            else:
                intersection_set_a.append(set(keys_5))
                intersection_set_f.append(set(keys_1))
        if not skip:
            if algo_name:
                i = i.replace('_' + algo_name, '')
                i = i.replace(algo_name + '_', '')
                i = i.replace(algo_name, '')
            i = i.replace('_all', '')
            union_set_1 = set.union(*intersection_set_1)
            union_set_5 = set.union(*intersection_set_5)
            union_set_a = set.union(*intersection_set_a)
            union_set_f = set.union(*intersection_set_f)
            intersection_set_1 = set.intersection(*intersection_set_1)
            intersection_set_5 = set.intersection(*intersection_set_5)
            intersection_set_a = set.intersection(*intersection_set_a)
            intersection_set_f = set.intersection(*intersection_set_f)
            if show:
                print(i)
                print('          >=5%/>=1%')
                print('intersect: %3d/%3d' % (len(intersection_set_5), len(intersection_set_1)))
                print('    union: %3d/%3d' % (len(union_set_5), len(union_set_1)))
                print()
            if store_in is not None:
                name = (i, methods[0], methods[1])
                if name not in store_in:
                    store_in[name] = {}
                store_in[name][i.replace('_', '\\_')] = [len(intersection_set_5), len(union_set_5)]
    return store_in


def generate_table_structure(pd, methods=None, store_in=None, algo_name=None, show=True, diagonal=False):
    if methods is None:
        methods = ['fanova', 'ablation', 'fanova_cut', 'local improvement analysis']
        combos = [['fanova', 'ablation'], ['fanova_cut', 'ablation'],
                  ['fanova', 'local improvement analysis'], ['fanova_cut', 'local improvement analysis'],
                  ['ablation', 'local improvement analysis']]
        if show:
            print('#'*180)
        store_in = {}
        for method in methods:
            if show:
                print(method)
            generate_table_structure(pd, [method], store_in, algo_name, show, diagonal)
            if show:
                print('#'*180)
        for combo in combos:
            if show:
                print(combo)
            generate_table_structure(pd, list(combo), store_in, algo_name, show, diagonal)
            if show:
                print('#'*180)
    elif len(methods) == 1:
        m = methods[0]
        index = pd[m].index
        store_in = _on_instance_sets(pd, m, index, store_in, algo_name, show, diagonal)
    else:
        index = pd[methods[0]].index
        store_in = _pairwise(pd, methods, index, store_in, algo_name, show)
    return store_in


def merge_data(result):
    tmp = sorted(list(result.keys()), key=lambda x: (x[1], x[2], x[0]))
    combos = [['fanova', 'ablation'], ['fanova_cut', 'ablation'],
              ['fanova', 'local improvement analysis'], ['fanova_cut', 'local improvement analysis'],
              ['ablation', 'local improvement analysis']]
    skip = False
    try:
        del tmp[tmp.index('fanova')]
    except ValueError:
        skip = True
    try:
        del tmp[tmp.index('ablation')]
    except ValueError:
        skip = True
    try:
        del tmp[tmp.index('fanova_cut')]
    except ValueError:
        skip = True
    try:
        del tmp[tmp.index('local improvement analysis')]
    except ValueError:
        skip = True
    df_dict = {}
    if not skip:
        multi = []
        all = None
        prev = tmp[0]
        for combo in combos:
            tap = np.array(tmp)
            indices = list(set.intersection(set(list(np.where(tap[:, 1] == combo[0]))[0]),
                                            set(list(np.where(tap[:, 2] == combo[1]))[0])))
            tap = list(tap[indices])
            for benchmark in tap:
                cols = ['cap_', 'cup_']
                if 'fanova' == benchmark[1]:
                    cols[0], cols[1] = cols[0] + 'f', cols[1] + 'f'
                elif 'ab' in benchmark[1]:
                    cols[0], cols[1] = cols[0] + 'a', cols[1] + 'a'
                else:
                    cols[0], cols[1] = cols[0] + 'c', cols[1] + 'c'
                if 'ab' in benchmark[2]:
                    cols[0], cols[1] = cols[0] + 'a', cols[1] + 'a'
                else:
                    cols[0], cols[1] = cols[0] + 'l', cols[1] + 'l'
                benchmark = tuple(benchmark)
                if all is None:
                    all = DataFrame.from_dict(result[benchmark], orient='index')
                    all.columns = cols
                    idx = list(map(lambda y: y[1], sorted(enumerate(list(all.index)), key=lambda x: x[1])))
                    all = all.loc[idx]
                else:
                    d = DataFrame.from_dict(result[benchmark], orient='index')
                    d.columns = cols
                    all = pd_concat([all, d])
            multi.append(all.copy())
            all = None
        df_dict['joint'] = pd_concat(multi, axis=1)
    for key in ['ablation', 'fanova', 'fanova_cut', 'local improvement analysis']:
        try:
            df = DataFrame.from_dict(result[key], orient='index')
            df.columns = ['cap', 'cup']
            idx = list(df.index)
            idx_0 = sorted(enumerate(list(map(lambda x: x.split(' / '), idx))), key=lambda x: (x[1][0], x[1][1]))
            idx_0 = list(map(lambda x: idx[x[0]], idx_0))
            df = df.loc[idx_0]
            df_dict[key] = df
        except KeyError:
            pass
    return df_dict


def create_latex_output(result):
    dfs = merge_data(result)
    skip = False
    if 'joint' not in dfs:
        skip = True
    if not skip:
        df = dfs['joint']
        tmp = list(df.keys())
        for idx, t in enumerate(tmp):
            t = t.split('_')
            t[0] = '$\\' + t[0]
            t[1] = [t[1][0], t[1][1]]
            t[1][0] = '{' + t[1][0] + '_{c5\\%}'
            t[1][1] = t[1][1] + '_{c5\\%}}$'
            t[1] = ''.join(t[1])
            t = '_'.join(t)
            tmp[idx] = t
        df.columns = tmp
        print('\\begin{table}[htbp]')
        print('\\centering')
        print(df.to_latex(escape=False, column_format='l' + 'c'*len(df.columns)))
        print('\\caption{%s}' % 'Comparison of fANOVA and ablation results on the same instance sets.')
        print('\\end{table}')
        print()
    for key in ['ablation', 'fanova', 'fanova_cut', 'local improvement analysis']:
        try:
            df = dfs[key]
            col = df['cap']
            coll = df['cup']
            tmp_idx = list(map(lambda x: x.split(' / '), list(col.index)))
            idx = list(map(lambda y: y[0], tmp_idx))
            idx.append(tmp_idx[-1][1])
            idx = sorted(list(set(idx)))
            data = []
            for i in idx:
                data.append([])
                for j in idx:
                    tmp = ' / '.join([i, j])
                    pmt = ' / '.join([j, i])
                    if tmp in list(col.index):
                        data[-1].append(col.loc[tmp])
                    elif pmt in list(col.index):
                        data[-1].append(coll.loc[pmt])
                    else:
                        data[-1].append('-')
            tmp_df = DataFrame.from_records(data, columns=idx)
            tmp_df.index = idx
            df = tmp_df
            if 'fanova' in key:
                key = key.replace('fanova', 'fANOVA')
            key = key.replace('_', '\\_')
            print('\\begin{table}[htbp]')
            print('\\centering')
            print(df.to_latex(escape=False, column_format='cl' + 'c' * len(idx), multirow=True))
            print('\\caption{Comparison of %s results across datasets. The lower triangle shows '
                  'the size of the union of all important parameters of both sets. The upper triangle shows '
                  'the size of the intersection of all important parameters.}' % key)
            print('\\end{table}')
            print()
        except KeyError:
            pass


def get_latex_outputs(algo_names, diagonal=False):
    rs = {}
    for algo in algo_names:
        v, _ = load_data_frames(algo, show=False)
        section = algo
        if 'clasp' in algo:
            if '_asp' in algo:
                algo = 'clasp'
            else:
                algo = 'clasp-3.0.4-p8'
        if 'clasp' in section:
            if '_asp' in section:
                section = 'clasp ASP'
            else:
                if 'random' in section:
                    section = 'clasp RAND'
                else:
                    section = 'clasp HAND'
        if section == 'cplex':
            section = 'CPLEX'
        elif section == 'satenstein':
            section = 'SATenstein'
        print('\\section{%s}' % section.replace('_', '\\_'))
        create_latex_output(generate_table_structure(v, algo_name=algo, show=False, diagonal=diagonal))
        print('\\clearpage')
        print()


def collect_all_dfs(algo_names, diagonal=False):
    rs = {}
    for algo in algo_names:
        v, _ = load_data_frames(algo)
        res = merge_data(generate_table_structure(v, algo_name=algo, diagonal=diagonal))
        rs[algo] = res
    return rs


def create_stats_df(algo_names, show=True, diagonal=False):
    all_ = collect_all_dfs(algo_names, diagonal=diagonal)
    res = []
    ros = []
    names = []
    for idx, algo in enumerate(algo_names):
        try:
            stat = [np.mean(all_[algo]['ablation']['cap'] / all_[algo]['ablation']['cup']),
                    np.std(all_[algo]['ablation']['cap'] / all_[algo]['ablation']['cup'])]
        except KeyError:
            stat.append(None)
            stat = [None, None]
        try:
            stat.append(np.mean(all_[algo]['fanova']['cap'] / all_[algo]['fanova']['cup']))
            stat.append(np.std(all_[algo]['fanova']['cap'] / all_[algo]['fanova']['cup']))
        except KeyError:
            stat.append(None)
            stat.append(None)
        try:
            stat.append(np.mean(all_[algo]['fanova_cut']['cap'] / all_[algo]['fanova_cut']['cup']))
            stat.append(np.std(all_[algo]['fanova_cut']['cap'] / all_[algo]['fanova_cut']['cup']))
        except KeyError:
            stat.append(None)
            stat.append(None)
        try:
            stat.append(np.mean(all_[algo][
                                    'local improvement analysis'][
                                    'cap'] / all_[algo]['local improvement analysis']['cup']))
            stat.append(np.std(all_[algo][
                                   'local improvement analysis'][
                                   'cap'] / all_[algo]['local improvement analysis']['cup']))
        except KeyError:
            stat.append(None)
            stat.append(None)

        try:
            stot = [np.mean(all_[algo]['joint']['cap_fa'] / all_[algo]['joint']['cup_fa']),
                    np.std(all_[algo]['joint']['cap_fa'] / all_[algo]['joint']['cup_fa'])]
        except KeyError:
            stot = [None, None]
        try:
            stot.append(np.mean(all_[algo]['joint']['cap_fl'] / all_[algo]['joint']['cup_fl']))
            stot.append(np.std(all_[algo]['joint']['cap_fl'] / all_[algo]['joint']['cup_fl']))
        except KeyError:
            stot.append(None)
            stot.append(None)
        try:
            stot.append(np.mean(all_[algo]['joint']['cap_ca'] / all_[algo]['joint']['cup_ca']))
            stot.append(np.std(all_[algo]['joint']['cap_ca'] / all_[algo]['joint']['cup_ca']))
        except KeyError:
            stot.append(None)
            stot.append(None)
        try:
            stot.append(np.mean(all_[algo]['joint']['cap_cl'] / all_[algo]['joint']['cup_cl']))
            stot.append(np.std(all_[algo]['joint']['cap_cl'] / all_[algo]['joint']['cup_cl']))
        except KeyError:
            stot.append(None)
            stot.append(None)
        try:
            stot.append(np.mean(all_[algo]['joint']['cap_al'] / all_[algo]['joint']['cup_al']))
            stot.append(np.std(all_[algo]['joint']['cap_al'] / all_[algo]['joint']['cup_al']))
        except KeyError:
            stot.append(None)
            stot.append(None)
        res.append(stat)
        ros.append(stot)
        if algo == 'cplex':
            names.append('CPLEX')
        elif algo == 'satenstein':
            names.append('SATenstein')
        elif 'clasp' in algo:
            if '_asp' in algo:
                algo = 'clasp ASP'
            else:
                if 'random' in algo:
                    algo = 'clasp RAND'
                else:
                    algo = 'clasp HAND'
            names.append(algo)
        else:
            names.append(algo)

    vs = DataFrame.from_records(ros, columns=[('fANOVA', 'vs. ablation', '$\mu$'),
                                              ('fANOVA', 'vs. ablation', '$\sigma$'),
                                              ('fANOVA', 'vs. LIN', '$\mu$'),
                                              ('fANOVA', 'vs. LIN', '$\sigma$'),
                                              ('fANOVA$\\_c$', 'vs. ablation', '$\mu$'),
                                              ('fANOVA$\\_c$', 'vs. ablation', '$\sigma$'),
                                              ('fANOVA$\\_c$', 'vs. LIN', '$\mu$'),
                                              ('fANOVA$\\_c$', 'vs. LIN', '$\sigma$'),
                                              ('ablation', 'vs. LIN', '$\mu$'),
                                              ('ablation', 'vs. LIN', '$\sigma$')
                                              ])
    df = DataFrame.from_records(res, columns=[
                                              ('Set vs. Set', 'ablation', '$\mu$'),
                                              ('Set vs. Set', 'ablation', '$\sigma$'),
                                              ('Set vs. Set', 'fANOVA', '$\mu$'),
                                              ('Set vs. Set', 'fANOVA', '$\sigma$'),
                                              ('Set vs. Set', 'fANOVA$\\_c$', '$\mu$'),
                                              ('Set vs. Set', 'fANOVA$\\_c$', '$\sigma$'),
                                              ('Set vs. Set', 'LIN', '$\mu$'),
                                              ('Set vs. Set', 'LIN', '$\sigma$')
                                              ])
    names = list(map(lambda x: x.replace('_', '\\_'), names))
    df *= 100
    vs *= 100
    df.index = names
    vs.index = names
    df.columns = pd.MultiIndex.from_tuples(df.columns, names=['', '', ''])
    vs.columns = pd.MultiIndex.from_tuples(vs.columns, names=['', '', ''])
    pd.options.display.float_format = '{:.2f}'.format
    if show:
        print()
        print('\\begin{table}[htbp]')
        print('\\centering')
        print(df.to_latex(escape=False, column_format='l|cc|cc|cc|cc|cc|cc', multicolumn_format='|c|'))
        print('\\caption{Comparison of fANOVA/ablation/LIN results across different instance sets.}')
        print('\\end{table}')
        print()
        print('\\begin{table}[htbp]')
        print('\\centering')
        print(vs.to_latex(escape=False, column_format='l|cc|cc|cc|cc|cc', multicolumn_format='|c|'))
        print('\\caption{Comparison of x and y results on the same instance sets.}')
        print('\\end{table}')
    return df, vs


def generate_all_possible_outputs(algo_names, diagonal=False):
    for algo in algo_names:
        jon_dirs(algo)
    df, vs = create_stats_df(algo_names, False, diagonal=diagonal)
    print('~*'*60)
    print('~*'*60)
    print('~*'*60)
    print()
    print()
    get_latex_outputs(algo_names, diagonal)
    print()
    print('\\section{%s}' % "Aggregated results")
    print('\\begin{table}[htbp]')
    print('\\centering')
    print(df.to_latex(escape=False, column_format='l|cc|cc|cc|cc|cc|cc', multicolumn_format='|c|'))
    print('\\caption{Comparison of fANOVA/ablation/LIN results across different instance sets.}')
    print('\\end{table}')
    print()
    print('\\begin{table}[htbp]')
    print('\\centering')
    print(vs.to_latex(escape=False, column_format='l|cc|cc|cc|cc|cc', multicolumn_format='|c|'))
    print('\\caption{Column 1 \\& 2: Comparison of x and y results on the same instance sets.}')
    print('\\end{table}')
    print('\\end{table}')



import itertools as it
import os
import pickle
import warnings
from collections import OrderedDict
from copy import deepcopy

import matplotlib as mpl
import numpy as np
import pandas as pd
from bokeh.models import Row, Panel, Tabs, Column
from tqdm import tqdm

from pimp.utils.bokeh_helpers import bokeh_boxplot, bokeh_heatmap_cat, bokeh_heatmap_num, save_and_show, \
    bokeh_multiline, bokeh_line_uncertainty

mpl.use('Agg')
from matplotlib import pyplot as plt

from smac.runhistory.runhistory import RunHistory
from ConfigSpace.configuration_space import Configuration
from ConfigSpace.util import impute_inactive_values
from ConfigSpace.hyperparameters import CategoricalHyperparameter, OrdinalHyperparameter, Constant

try:
    from fanova import fANOVA as fanova_pyrfr
    from fanova.visualizer import Visualizer
    from fanova.__version__ import __version__ as fanova_version
except ImportError:
    warnings.simplefilter('always', ImportWarning)
    warnings.warn('\n{0}\n{0}{1}{0}\n{0}'.format('!'*120,
                                     '\nfANOVA is not installed in your environment. To install it please run '
                                     '"git+http://github.com/automl/fanova.git@master"\n'))

from pimp.evaluator.base_evaluator import AbstractEvaluator


__author__ = "Andre Biedenkapp"
__copyright__ = "Copyright 2016, ML4AAD"
__license__ = "3-clause BSD"
__maintainer__ = "Andre Biedenkapp"
__email__ = "biedenka@cs.uni-freiburg.de"


class fANOVA(AbstractEvaluator):

    def __init__(self, scenario, cs, model, to_evaluate: int, runhist: RunHistory, rng,
                 n_pairs=5, minimize=True, pairwise=True, preprocessed_X=None, preprocessed_y=None,
                 incumbents=None, **kwargs):
        """
        Handler to fANOVA module.

        Parameters
        ----------
        scenario: Scenario
            scenario with information about run_objective
        cs: ConfigSpace
            configuration space of scenario to be analyzed
        model: empirical performance model
            TODO
        to_evaluate: int
            number of parameters to be plotted
        runhist: RunHistory
            TODO
        rng: RandomNumberGenerator
            rng
        n_pairs: int
            how many (most important) parameters should be plotted for pairwise
            marginals
        minimize: boolean
            whether optimum is min or max
        pairwise: boolean
            plot pairwise marginals
        preprocessed_X/Y: data
            preprocessed data to be reused if model is already trained on data
            without instance_features
        incumbents: List[Configuration] or Configuration
            one or multiple incumbents to be marked in plots
        """
        super().__init__(scenario, cs, model, to_evaluate, rng, **kwargs)
        self.name = 'fANOVA'
        self.logger = 'pimp.' + self.name

        # Turn all Constants into Categoricals (fANOVA cannot handle Constants)
        self.cs_contained_constant = False
        # if any([isinstance(hp, Constant) for hp in self.cs.get_hyperparameters()]):
        #     self.logger.debug("Replacing configspace's hyperparameter Constants by one-value Categoricals.")
        #     new_hyperparameters = [CategoricalHyperparameter(hp.name, [hp.value]) if isinstance(hp, Constant)
        #                            else hp for hp in self.cs.get_hyperparameters()]
        #     self.cs = ConfigurationSpace()
        #     self.cs.add_hyperparameters(new_hyperparameters)
        #     self.cs_contained_constant = True

        # This way the instance features in X are ignored and a new forest is constructed
        if self.model.instance_features is None:
            self.logger.info('No marginalization necessary')
            if preprocessed_X is not None and preprocessed_y is not None:
                self._X = preprocessed_X
                self._y = preprocessed_y
            else:
                self.logger.info('Preprocessing X')
                self._X = deepcopy(self.X)
                self._y = deepcopy(self.y)
                for c_idx, config in enumerate(self.X):
                    # print("{}/{}".format(c_idx, len(self.X)))
                    for p_idx, param in enumerate(self.cs.get_hyperparameters()):
                        if not isinstance(param, (CategoricalHyperparameter, OrdinalHyperparameter, Constant)):
                            # getting the parameters out of the hypercube setting as used in smac runhistory
                            self._X[c_idx][p_idx] = param._transform(self.X[c_idx][p_idx])
        else:
            self._preprocess(runhist)
        cutoffs = (-np.inf, np.inf)
        if minimize:
            cutoffs = (-np.inf, self.model.predict_marginalized_over_instances(
                np.array([impute_inactive_values(self.cs.get_default_configuration()).get_array()]))[0].flatten()[0]
                       )
        elif minimize is False:
            cutoffs = (self.model.predict_marginalized_over_instances(
                np.array([impute_inactive_values( self.cs.get_default_configuration()).get_array()]))[0].flatten()[0],
                       np.inf)
        if fanova_version > '2.0.17':
            self._X = pd.DataFrame(self._X, columns=self.cs.get_hyperparameter_names())
        self.evaluator = fanova_pyrfr(X=self._X, Y=self._y.flatten(), config_space=self.cs,
                                      seed=self.rng.randint(2**31-1), cutoffs=cutoffs)
        self.n_most_imp_pairs = n_pairs
        self.num_single = None
        self.pairwise = pairwise
        self.evaluated_parameter_importance_uncertainty = OrderedDict()
        self.incumbents = incumbents

    def _preprocess(self, runhistory):
        """
        Method to marginalize over instances such that fANOVA can determine the parameter importance without
        having to deal with instance features.
        :param runhistory: RunHistory that knows all configurations that were run. For all these configurations
                           we have to marginalize away the instance features with which fANOVA will make it's
                           predictions
        """
        self.logger.info('PREPROCESSING PREPROCESSING PREPROCESSING PREPROCESSING PREPROCESSING PREPROCESSING')
        self.logger.info('Marginalizing away all instances!')
        configs = runhistory.get_all_configs()
        if self.cs_contained_constant:
            configs = [Configuration(self.cs, vector=c.get_array()) for c in configs]
        X_non_hyper, X_prime = [], []
        for config in configs:
            config = impute_inactive_values(config).get_array()
            X_prime.append(config)
            X_non_hyper.append(config)
            for idx, param in enumerate(self.cs.get_hyperparameters()):
                if not (isinstance(param, CategoricalHyperparameter) or
                        isinstance(param, Constant)):
                    X_non_hyper[-1][idx] = param._transform(X_non_hyper[-1][idx])
        X_non_hyper = np.array(X_non_hyper)
        X_prime = np.array(X_prime)
        y_prime = np.array(self.model.predict_marginalized_over_instances(X_prime)[0])
        self._X = X_non_hyper
        self._y = y_prime
        self.logger.info('Size of training X after preprocessing: %s' % str(self.X.shape))
        self.logger.info('Size of training y after preprocessing: %s' % str(self.y.shape))
        self.logger.info('Finished Preprocessing')

    def _get_label(self, run_obj):
        if run_obj == 'runtime':
            label = 'runtime [sec]'
        elif run_obj == 'quality':
            label = 'cost'
        else:
            label = '%s' % self.scenario.run_obj
        return label

    def plot_result(self, name='fANOVA', show=True):
        if not os.path.exists(name):
            os.mkdir(name)

        vis = Visualizer(self.evaluator, self.cs, directory=name, y_label=self._get_label(self.scenario.run_obj))
        self.logger.info('Getting Marginals!')
        pbar = tqdm(range(self.to_evaluate), ascii=True, disable=not self.verbose)
        for i in pbar:
            plt.close('all')
            plt.clf()
            param = list(self.evaluated_parameter_importance.keys())[i]
            # Plot once in log, once linear
            for mode in [(True, '_log'), (False, '')]:
                outfile_name = os.path.join(name, param.replace(os.sep, "_") + mode[1] + ".png")
                # The try/except clause is only for back-compatibility with fanova <= 2.0.11
                try:
                    vis.plot_marginal(self.cs.get_idx_by_hyperparameter_name(param), log_scale=mode[0], show=False, incumbents=self.incumbents)
                except TypeError:
                    self.logger.debug("Plotting incumbents not supported by fanova < 2.0.12")
                    vis.plot_marginal(self.cs.get_idx_by_hyperparameter_name(param), log_scale=mode[0], show=False)
                fig = plt.gcf()
                fig.savefig(outfile_name)
                plt.close('all')
                plt.clf()
            if show:
                plt.show()
            pbar.set_description('Creating fANOVA plot: {: <.30s}'.format(outfile_name.split(os.path.sep)[-1]))
        if self.pairwise:
            self.logger.info('Plotting Pairwise-Marginals!')
            most_important_ones = list(self.evaluated_parameter_importance.keys())[
                                  :min(self.num_single, self.n_most_imp_pairs)]
            vis.create_most_important_pairwise_marginal_plots(most_important_ones)
            try:
                vis.create_most_important_pairwise_marginal_plots(most_important_ones)
            except TypeError as err:
                self.logger.debug(err, exc_info=1)
                self.logger.warning('Could not create pairwise plots!')
            plt.close('all')

    def plot_bokeh(self, plot_name=None, show_plot=False, plot_pairwise="most_important"):
        """
        Plot single and pairwise margins in bokeh-plot. Single margins are always plotted (not expensive), pairwise can
        be configured by argument.

        Parameters
        ----------
        plot_name: str
            path where to store the plot, None to not save it
        show_plot: bool
            whether or not to open plot in standard browser
        plot_pairwise: str
            choose from ["none", "most_important", "all"] where "most_important" relies on the fanova module to decide
            what that means

        Returns
        -------
        layout: bokeh.models.Column
            bokeh plot (can be used in notebook or comparted with components)
        """
        vis = Visualizer(self.evaluator, self.cs, directory='.', y_label=self._get_label(self.scenario.run_obj))

        ####################
        # Single marginals #
        ####################
        plots_single = []
        params = list(self.evaluated_parameter_importance.keys())
        pbar = tqdm(deepcopy(params), ascii=True, disable=not self.verbose)
        for param_name in pbar:
            # Try and except pairwise importances that are also saved in evaluated_parameter_importance...
            try:
                param = self.cs.get_hyperparameter(param_name)
            except KeyError as err:
                self.logger.debug(err, exc_info=1)
                continue

            pbar.set_description('Plotting fANOVA (in bokeh) for %s' % param_name)

            incumbents = []
            if not self.incumbents is None:
                incumbents = self.incumbents.copy() if isinstance(self.incumbents, list) else [self.incumbents]
            values = [c[param_name] for c in incumbents if param_name in c and c[param_name] is not None]

            if isinstance(param, (CategoricalHyperparameter, Constant)):
                labels = param.choices if isinstance(param, CategoricalHyperparameter) else str(param)
                mean, std = vis.generate_marginal(param_name)
                inc_indices = [labels.index(val) for val in values]

                p = bokeh_boxplot(labels, mean, std,
                                  x_label=param.name,
                                  y_label="runtime [sec]" if self.scenario.run_obj == "runtime" else "cost",
                                  runtime=self.scenario.run_obj=="runtime",
                                  inc_indices=inc_indices)

            else:
                mean, std, grid = vis.generate_marginal(param_name, 100)
                mean, std = np.asarray(mean), np.asarray(std)
                log_scale = param.log or (np.diff(grid).std() > 0.000001)
                inc_indices = [(np.abs(np.asarray(grid) - val)).argmin() for val in values]

                p = bokeh_line_uncertainty(grid, mean, std, log_scale,
                                           x_label=param.name,
                                           y_label="runtime [sec]" if self.scenario.run_obj == "runtime" else "cost",
                                           inc_indices=inc_indices)

            plots_single.append(Panel(child=Row(p), title=param_name))

        ######################
        # Pairwise marginals #
        ######################
        if plot_pairwise == "all":
            combis = list(it.combinations(self.cs.get_hyperparameters(), 2))
        elif plot_pairwise == "most_important":
            most_important_ones = list(self.evaluated_parameter_importance.keys())[
                                  :min(self.num_single, self.n_most_imp_pairs)]
            most_important_pairwise_marginals = vis.fanova.get_most_important_pairwise_marginals(
                                                params=most_important_ones)
            combis = [(self.cs.get_hyperparameter(name1), self.cs.get_hyperparameter(name2)) for name1, name2 in
                      most_important_pairwise_marginals]
        elif plot_pairwise == "none":
            combis = []
        else:
            raise ValueError("{} not a valid set of pairwise plots to generate...".format(plot_pairwise))

        plots_pairwise = []
        pbar = tqdm(deepcopy(combis), ascii=True, disable=not self.verbose)
        for p1, p2 in pbar:
            pbar.set_description('Plotting pairwise fANOVA (in bokeh) for %s & %s' % (p1.name, p2.name))
            first_is_cat = isinstance(p1, CategoricalHyperparameter)
            second_is_cat = isinstance(p2, CategoricalHyperparameter)
            # There are essentially three cases / different plots:
            # First case: both categorical -> heatmap
            if first_is_cat or second_is_cat:
                choices, zz = vis.generate_pairwise_marginal((p1.name, p2.name), 20)
                # Working with pandas makes life easier
                data = pd.DataFrame(zz, index=choices[0], columns=choices[1])
                # Setting names for rows and columns and make categoricals strings
                data.index.name, data.columns.name = p1.name, p2.name
                data.index = data.index.astype(str) if first_is_cat else data.index
                data.columns = data.columns.astype(str) if second_is_cat else data.columns
                if first_is_cat and second_is_cat:
                    p = bokeh_heatmap_cat(data, p1.name, p2.name)
                else:
                    # Only one of them is categorical -> create multi-line-plot
                    cat_choices = p1.choices if first_is_cat else p2.choices
                    # We want categorical values be represented by columns:
                    if not second_is_cat:
                        data = data.transpose()
                    # Find y_min and y_max BEFORE resetting index (otherwise index max obscure the query)
                    y_limits = (data.min().min(), data.max().max())
                    x_limits = (p1.lower if second_is_cat else p2.lower, p1.upper if second_is_cat else p2.upper)
                    # We want the index as a column (for plotting on x-axis)
                    data = data.reset_index()
                    p = bokeh_multiline(data, x_limits, y_limits,
                                        p1.name if second_is_cat else p2.name,
                                        cat_choices,
                                        y_label="runtime [sec]" if self.scenario.run_obj == "runtime" else "cost",
                                        z_label=p1.name if first_is_cat else p2.name,
                                        )

            else:
                # Third case: both continous
                grid, zz = vis.generate_pairwise_marginal((p1.name, p2.name), 20)
                data = pd.DataFrame(zz, index=grid[0], columns=grid[1])
                data.index.name, data.columns.name = p1.name, p2.name
                p = bokeh_heatmap_num(data, p1.name, p2.name, p1.log, p2.log)

            plots_pairwise.append(Panel(child=Row(p), title=" & ".join([p1.name, p2.name])))

        # Putting both together
        tabs_single = Tabs(tabs=[*plots_single])
        if len(plots_pairwise) > 0:
            tabs_pairwise = Tabs(tabs=[*plots_pairwise])
            layout = Column(tabs_single, tabs_pairwise)
        else:
            layout = Column(tabs_single)

        # Save and show...
        save_and_show(plot_name, show_plot, layout)

        return layout

    def run(self) -> OrderedDict:
        try:
            params = self.cs.get_hyperparameters()

            tmp_res = []
            for idx, param in enumerate(params):
                imp = self.evaluator.quantify_importance([idx])[(idx, )]
                self.logger.debug('{:>02d} {:<30s}: {:>02.4f}' .format(
                    idx, param.name, imp['total importance']))
                tmp_res.append(imp)

            tmp_res_sort_keys = [i[0] for i in sorted(enumerate(tmp_res), key=lambda x:x[1]['total importance'], reverse=True)]
            self.logger.debug(tmp_res_sort_keys)
            count = 0
            for idx in tmp_res_sort_keys:
                if count >= self.to_evaluate:
                    break
                self.logger.info('{:>02d} {:<30s}: {:>02.4f}'.format(idx, params[idx].name, tmp_res[idx]['total importance']))
                self.evaluated_parameter_importance[params[idx].name] = tmp_res[idx]['total importance']
                try:
                    self.evaluated_parameter_importance_uncertainty[params[idx].name] = tmp_res[idx]['total std']
                except KeyError as e:
                    self.logger.debug("std not available yet for this fanova version")
                count += 1
            self.num_single = len(list(self.evaluated_parameter_importance.keys()))
            if self.pairwise:
                self.logger.info(
                    'Computing most important pairwise marginals using at most'
                    ' the %d most important ones.' % min(self.n_most_imp_pairs, self.num_single))
                pairs = [x for x in it.combinations(list(self.evaluated_parameter_importance.keys())[:self.n_most_imp_pairs], 2)]
                for pair in pairs:
                    imp = self.evaluator.quantify_importance(pair)[pair]
                    try:
                        mean, std = imp['individual importance'], imp['individual std']
                        self.evaluated_parameter_importance_uncertainty[str(list(pair))] = std
                    except KeyError as e:
                        self.logger.debug("std not available yet for this fanova version")
                        mean= imp['individual importance']
                    self.evaluated_parameter_importance[str(list(pair))] = mean
                    a, b = pair
                    if len(a) > 13:
                        a = str(a)[:5] + '...' + str(a)[-5:]
                    if len(b) > 13:
                        b = str(b)[:5] + '...' + str(b)[-5:]
                    self.logger.info('{:>02d} {:<30s}: {:>02.4f}'.format(-1, a + ' <> ' + b, mean))
            all_res = {'imp': self.evaluated_parameter_importance,
                       'order': list(self.evaluated_parameter_importance.keys())}
            return all_res
        except ZeroDivisionError:
            with open('fANOVA_crash_data.pkl', 'wb') as fh:
                pickle.dump([self.X, self.y, self.cs], fh)
            raise Exception('fANOVA crashed with a "float division by zero" error. Dumping the data to disk')
